/*---------------------------------------------------------------------------*/
/* INDEXED 2D ARRAYS */

extern int cp40_alloc_2d_index_b(cp40_byte_t ***index_ptr, cp40_byte_t *array,
				int xmin, int xmax, int ymin, int ymax);
extern int cp40_alloc_2d_index_s(short ***index_ptr, short *array,
				int xmin, int xmax, int ymin, int ymax);
extern int cp40_alloc_2d_index_i(int ***index_ptr, int *array,
				int xmin, int xmax, int ymin, int ymax);
extern int cp40_alloc_2d_index_l(long ***index_ptr, long *array,
				int xmin, int xmax, int ymin, int ymax);
extern int cp40_alloc_2d_index_f(float ***index_ptr, float *array,
				int xmin, int xmax, int ymin, int ymax);
extern int cp40_alloc_2d_index_d(double ***index_ptr, double *array,
				int xmin, int xmax, int ymin, int ymax);
extern int cp40_alloc_2d_index_c(cp40_fcomplex_t ***index_ptr,
				cp40_fcomplex_t *array,
				int xmin, int xmax, int ymin, int ymax);
extern int cp40_alloc_2d_index_z(cp40_dcomplex_t ***index_ptr,
				cp40_dcomplex_t *array,
				int xmin, int xmax, int ymin, int ymax);
/*	Allocate and instanciate array of pointers for addressing 2D array
	at base address ARRAY.  A standard result, CP40_OK or CP40_ERROR is
        returned.  Assuming INDEX_PTR=&INDEX, then
	    INDEX[Y][X]
	is a faster indirection than
	    ARRAY[(X-XMIN) + (XMAX-XMIN+1) * (Y-YMIN)]
	for XMIN<=X<=XMAX and YMIN<=Y<=YMAX. */

extern void cp40_free_2d_index_b(cp40_byte_t **index,
				int xmin, int xmax, int ymin, int ymax);
extern void cp40_free_2d_index_s(short **index,
				int xmin, int xmax, int ymin, int ymax);
extern void cp40_free_2d_index_i(int **index,
				int xmin, int xmax, int ymin, int ymax);
extern void cp40_free_2d_index_l(long **index,
				int xmin, int xmax, int ymin, int ymax);
extern void cp40_free_2d_index_f(float **index,
				int xmin, int xmax, int ymin, int ymax);
extern void cp40_free_2d_index_d(double **index,
				int xmin, int xmax, int ymin, int ymax);
extern void cp40_free_2d_index_c(cp40_fcomplex_t **index,
				int xmin, int xmax, int ymin, int ymax);
extern void cp40_free_2d_index_z(cp40_dcomplex_t **index,
				int xmin, int xmax, int ymin, int ymax);
/*	Free array of pointers allocated by the corresponding
	cp40_alloc_2d_index_? routine. */

extern int cp40_alloc_2d_index(void *index_ptr, unsigned elsize, void *array,
			      int xmin, int xmax, int ymin, int ymax);
extern void cp40_free_2d_index(void *index, unsigned elsize,
			      int xmin, int xmax, int ymin, int ymax);
/*	Similar to typed alloc/free 2D index routines but designed for
	more general cases: only the element size (in bytes) ELSIZE is
	required.   The usage is examplified in the following piece of
	code (CTYPE is any primary or typedef'ed C data type):
	  CTYPE **index, array[(XMAX-XMIN+1) * (YMAX-YMIN+1)];
	  if (cp40_alloc_2d_index(&pixel, sizeof(CTYPE), array,
	                       XMIN, XMAX, YMIN, YMAX) != CP40_OK) {
	    error
          }
	  ...
	  cp40_free_2d_index(index, sizeof(CTYPE), XMIN, XMAX, YMIN, YMAX);
*/

extern int  cp40_alloc_matrix(void *index_ptr, unsigned elsize,
			     int xmin, int xmax, int ymin, int ymax);
extern void cp40_free_matrix(void *index, unsigned elsize,
			    int xmin, int xmax, int ymin, int ymax);
/*	Similar to cp40_alloc_2d_index/cp40_free_2d_index but also
	allocate/free memory for array data.  The allocated matrix is
	filled with zeroes. */

static int alloc_2d(void *index_ptr, unsigned elsize, void *ptr,
		   int xmin, int xmax, int ymin, int ymax);

static int alloc_2d(void *index_ptr, unsigned elsize, void *ptr,
		   int xmin, int xmax, int ymin, int ymax)
{
  int width= xmax-xmin+1;
  int height= ymax-ymin+1;
  int pitch= width*elsize;
  int y;
  char **index, *array= ptr;

  /* Check arguments. */
  if (elsize<=0) {
    return cp40_error("bad element size");
  }
  if (width<=0 || height<=0) {
    return cp40_error("bad boundaries");
  }

  /* Allocate memory. */
  CP40_ALLOC_AT(index, height);
  if (array) {
    if (! index) {
      cp40_error("insufficient memory for 2D index");
    }
  } else {
    CP40_ALLOC_AT(array, height*pitch);
    if (! array || ! index) {
      CP40_FREE_MAYBE(index);
      CP40_FREE_MAYBE(array);
      return cp40_error("insufficient memory for indexed matrix");
    }
    memset(array, 0, height*pitch);
  }

  /* Instanciate index. */
  index-= ymin;
  index[ymin]= array - xmin*elsize;
  for (y=ymin ; y<ymax ; y++) {
    index[y+1]= index[y] + pitch;
  }
  *(char ***)index_ptr= index;
  return CP40_OK;
}

#undef FUNCTION
#define FUNCTION(NAME, CTYPE)						      \
int NAME(CTYPE ***index_ptr, CTYPE *array,				      \
	 int xmin, int xmax, int ymin, int ymax)			      \
{ return cp40_alloc_2d_index(index_ptr, sizeof(CTYPE), array,		      \
			  xmin, xmax, ymin, ymax); }
FUNCTION(cp40_alloc_2d_index_b, unsigned char)
FUNCTION(cp40_alloc_2d_index_s, short)
FUNCTION(cp40_alloc_2d_index_i, int)
FUNCTION(cp40_alloc_2d_index_l, long)
FUNCTION(cp40_alloc_2d_index_f, float)
FUNCTION(cp40_alloc_2d_index_d, double)
FUNCTION(cp40_alloc_2d_index_c, cp40_fcomplex_t)
FUNCTION(cp40_alloc_2d_index_z, cp40_dcomplex_t)

#undef FUNCTION
#define FUNCTION(NAME, CTYPE)						      \
void NAME(CTYPE **index, int xmin, int xmax, int ymin, int ymax)	      \
{ cp40_free_2d_index(index, sizeof(CTYPE), xmin, xmax, ymin, ymax); }
FUNCTION(cp40_free_2d_index_b, unsigned char)
FUNCTION(cp40_free_2d_index_s, short)
FUNCTION(cp40_free_2d_index_i, int)
FUNCTION(cp40_free_2d_index_l, long)
FUNCTION(cp40_free_2d_index_f, float)
FUNCTION(cp40_free_2d_index_d, double)
FUNCTION(cp40_free_2d_index_c, cp40_fcomplex_t)
FUNCTION(cp40_free_2d_index_z, cp40_dcomplex_t)

int cp40_alloc_2d_index(void *index_ptr, unsigned elsize, void *array,
		     int xmin, int xmax, int ymin, int ymax)
{
  if (! array) return cp40_error("bad array address");
  return alloc_2d(index_ptr, elsize, array, xmin, xmax, ymin, ymax);
}

void cp40_free_2d_index(void *index, unsigned elsize,
		     int xmin, int xmax, int ymin, int ymax)
{
  if (elsize>0 && xmax>=xmin && ymax>=ymin) {
    char **row= index;
    free(&row[ymin]);
  }
}

int cp40_alloc_matrix(void *index_ptr, unsigned elsize,
		   int xmin, int xmax, int ymin, int ymax)
{
  return alloc_2d(index_ptr, elsize, NULL, xmin, xmax, ymin, ymax);
}

void cp40_free_matrix(void *index, unsigned elsize,
		     int xmin, int xmax, int ymin, int ymax)
{
  if (elsize>0 && xmax>=xmin && ymax>=ymin) {
    char **row= index;
    free(&row[ymin][xmin*elsize]);
    free(&row[ymin]);
  }
}

/*---------------------------------------------------------------------------*/
/* SMALL BLOCKS OF MEMORY */

/* CP40 provides a faster (than cp40_mem_alloc/cp40_mem_free or
   cp40_alloc/cp40_free) way of handling small units of memory of a given size
   (adapted from David Munro's `defmem.c' in Yorick-1.4). */

/* The `cp40_mem_block_t' structure defines the size of memory units and blocks
   (chunk of units allocated by `cp40_alloc') and is also used to keep track of
   the list of allocated blocks.  Each allocated block has a small header
   (rounded up to CP40_ALIGN bytes and used to store the address of the next
   allocated block) followed by UNITS_PER_BLOCK units of UNIT_SIZE bytes (each
   unit is aligned on a multiple of CP40_ALIGN bytes so the effective unit size
   is UNIT_SIZE rounded up to CP40_ALIGN bytes).  */

typedef struct _cp40_mem_block cp40_mem_block_t;

struct _cp40_mem_block {
  void *unit_list;        /* Allocate new block when reaches 0.  Must be
			     initially set to NULL. */
  void *block_list;       /* List of blocks kept using first unit of each
			     block.  Must be initially set to NULL. */
  size_t unit_size;       /* Byte size of units (must be >= 1 byte, effective
			     unit_size is rounded up to CP40_ALIGN bytes). */
  size_t units_per_block; /* Number of units per memory block (must be at
			     least one). */
  /* Note: the restrictions about members values must be strictly respected
     otherwise the memory block manager will fail.  You can use
     `cp40_mem_init_unit()' to that end. */
};

/* Macro to round up size of memory units in cp40_unit_descr_t. */
#define CP40_MEM_ROUND_UNIT_SIZE(s)   CP40_ROUND_UP(s, CP40_ALIGN)

extern void cp40_mem_init_unit(cp40_mem_block_t *mb, size_t unit_size,
			      size_t units_per_block);
/*	Initialize MB appropriately.  Must be called only once before the
	first `cp40_mem_next_unit(MB)'.  MB->unit_size and MB->UNITS_PER_BLOCK
	are both set to at least one.  Effective UNIT_SIZE is rounded up to
	CP40_ALIGN bytes. */

extern void *cp40_mem_next_unit(cp40_mem_block_t *mb);
/*	Allocate memory unit managed by MB and with size: BLOCK->unit_size.
	In case of error, NULL is returned (error message can be retrieved by
	cp40_error_info).  It is the caller responsibility to free the memory
	with `cp40_mem_free_unit'. */

extern void cp40_mem_free_unit(cp40_mem_block_t *mb, void *unit);
/*	Free memory unit, UNIT must be the result of an `cp40_unit_next(MB)'
	call.  In fact, UNIT is returned to the pool of free units managed by
	MB, use `cp40_mem_free_all_units' to effectively free *ALL* units
	managed by MB.  */

extern void cp40_mem_free_all_units(cp40_mem_block_t *mb);
/*	Free all units managed by MB. */

/*---------------------------------------------------------------------------*/
/* MEMORY BLOCKS */

extern void  cp40_mem_free_all_units(cp40_mem_block_t *block);
extern void  cp40_mem_free_unit(cp40_mem_block_t *block, void *unit);
extern void *cp40_mem_next_unit(cp40_mem_block_t *block);

static void *new_block(cp40_mem_block_t *mb);

void cp40_mem_init_unit(cp40_mem_block_t *mb, size_t unit_size,
		       size_t units_per_block)
{
  mb->unit_list= NULL;
  mb->block_list= NULL;
  mb->unit_size= CP40_MAX(unit_size, 1);
  mb->units_per_block= CP40_MAX(units_per_block, 1);
}

void *cp40_mem_next_unit(cp40_mem_block_t *mb)
{
  void *unit= mb->unit_list;

  if (unit) {
    /* (void**)<->(void*) typecast in next line is equivalent to
       list->next if list is a struct List {struct List *next;}... */
    mb->unit_list= *(void **)unit;
    return unit;
  }
  return new_block(mb);
}

void cp40_mem_free_unit(cp40_mem_block_t *mb, void *unit)
{
  *(void **)unit= mb->unit_list;
  mb->unit_list= unit;
}

void cp40_mem_free_all_units(cp40_mem_block_t *mb)
{
  void *block, *next= mb->block_list;
  while ((block= next) != NULL) {
    next= *(void **)block;
    free(block);
  }
}

#undef BLOCK_HEADER_SIZE
#define BLOCK_HEADER_SIZE CP40_ROUND_UP(sizeof(void *), CP40_ALIGN)

static void *new_block(cp40_mem_block_t *mb)
{
  void *block;
  int block_size, unit_size, units_per_block;

  /* Allocate a new block and insert it in the chain of allocated blocks. */
  if ((unit_size= CP40_ROUND_UP(mb->unit_size, CP40_ALIGN)) < sizeof(void *)
      || (units_per_block= mb->units_per_block) < 1) {
    cp40_error("bad memory block definition");
    return NULL;
  }
  block_size= BLOCK_HEADER_SIZE + unit_size * units_per_block;
  if (! (block= malloc(block_size))) {
    cp40_error("not enough memory for a new block of units");
    return NULL;
  }
  *(void **)block= mb->block_list;
  mb->block_list= block;

  /* Adavnce to 1st unit in new block. */
  block= (void *)((char *)block + BLOCK_HEADER_SIZE);

  /* Chain other allocated units. */
  if (units_per_block > 1) {
    char *next= (char *)block + unit_size;
    char *last= (char *)block + unit_size*(units_per_block-1);
    char *unit;
    while ((unit= next) < last) {
      next+= unit_size;
      *(void **)unit= next;
    }
    *(void **)last= mb->unit_list;
    mb->unit_list= (char *)block + unit_size;
  }
  return block;
}

/*---------------------------------------------------------------------------*/
