/*
 * cp40.h --
 *
 *	Declarations of externally visible resources of the CP40 library.
 *	The numerous comments throughout this file are intended to provide
 *	a basic documentation for the use of the CP40 library.
 *
 *-----------------------------------------------------------------------------
 *
 *	Copyright (C) 1995-2006 Eric Thibaut.
 *
 *	This file is part of CP40soft.
 *
 *	CP40soft is free software; you can redistribute it and/or modify it
 *	under the terms of the GNU General Public License as published by
 *	the Free Software Foundation; either version 2 of the License, or
 *	(at your option) any later version.
 *
 *	CP40soft is distributed in the hope that it will be useful, but
 *	WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *	General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with CP40soft; if not, write to the Free Software Foundation,
 *	Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
 *
 *-----------------------------------------------------------------------------
 *
 * History:
 *	$Id$
 *	$Log$
 */

#ifndef _CP40_H
#define _CP40_H 1

/* Standard header files needed. */
#ifndef BUFSIZ
# include <stdio.h>
#endif

/*---------------------------------------------------------------------------*/

/* Define some macros to get rid of some GNU extensions when not compiling
   with GCC. */
#if ! (defined(__GNUC__) && __GNUC__ > 1)
#   define __attribute__(x)
#   define __inline__
#   define __FUNCTION__        ""
#   define __PRETTY_FUNCTION__ ""
#endif

/* NULL pointer (I hope this definition is not specific to CP40 ;-). */
#ifndef NULL
# define NULL ((void *)0)
#endif

/* Define CP40_DEBUG to non-zero to replace fast/in-lined/macros
   functions by slower version that may call abort() in case of error. */
#ifndef CP40_DEBUG
# define CP40_DEBUG 0
#endif

/* Suitable alignement for any data types. Alignement to sizeof(double) should
   be OK for all CPU's (some 32-bit machine tolerate 4-byte address
   justification but this may penalize manipulation of double precision
   value's, e.g. on a Pentium). */
#define CP40_ALIGN sizeof(double)

/* _CP40_BEGIN_DECLS should be used at the beginning of your declarations,
   so that C++ compilers don't mangle their names.  Use _CP40_END_DECLS at
   the end of C declarations. */
#ifdef __cplusplus
# define _CP40_BEGIN_DECLS extern "C" {
# define _CP40_END_DECLS }
#else
# define _CP40_BEGIN_DECLS /* empty */
# define _CP40_END_DECLS /* empty */
#endif

/*---------------------------------------------------------------------------*/
/* USEFUL MACROS */

/* CP40_STRINGIFY takes an argument and wraps it in "" (double quotation
   marks), CP40_JOIN joins two arguments. */
#ifdef __STDC__
# define CP40_STRINGIFY(x)   #x
# define CP40_JOIN(a,b)      a##b
# define CP40_JOIN2(a,b)     a##b
# define CP40_JOIN3(a,b,c)   a##b##c
# define CP40_JOIN4(a,b,c,d) a##b##c##d
#else
# define CP40_STRINGIFY(x)   "x"
# define CP40_JOIN(a,b)      a/**/b
# define CP40_JOIN2(a,b)     a/**/b
# define CP40_JOIN3(a,b,c)   a/**/b/**/c
# define CP40_JOIN4(a,b,c,d) a/**/b/**/c/**/d
#endif

/* CP40_ASSERT(expr) check if expr is true, otherwise printout error message
   and abort. */
#define CP40_ASSERT(expr) if (!(expr)) cp40_assert_fail(CP40_STRINGIFY(expr), \
					                 __FILE__, __LINE__)

#define CP40_ABS(a)   ((a)>=0?(a):-(a))
/*	Computes absolute value. */

#define CP40_MIN(a,b) ((a)<=(b)?(a):(b))
#define CP40_MAX(a,b) ((a)>=(b)?(a):(b))
/*	Computes min/max values. */

#define CP40_HOW_MANY(nelem, grain) (((nelem)+((grain)-1))/(grain))
/*	Computes minimal number of chunks with GRAIN elements needed to store
	NELEM elements: */

#define CP40_ROUND_UP(nelem, grain) (CP40_HOW_MANY(nelem, grain)*(grain))
/*	Returns NELEM elements rounding up to a multiple of GRAIN elements. */

#define CP40_OFFSET_OF(struc, member) ((size_t) &((struc *)0)->member)
/*	Offset (in bytes) of MEMBER in structure STRUC. */

#define CP40_REALLOC(ptr, what, how_many) \
          ((what *)((ptr)?realloc(ptr,(how_many)*sizeof(what))\
		         :malloc((how_many)*sizeof(what))))
/*	(Re)allocate memory for HOW_MANY number elements of type WHAT at PTR,
	PTR can be NULL. */

#define CP40_ALLOC_ONE(what)       ((what*)malloc(sizeof(what)))
/*	Allocate memory for one element of type WHAT. */

#define CP40_ALLOC_MANY(what, how_many) \
          ((what*)malloc((how_many)*sizeof(what)))
/*	Allocate memory for HOW_MANY number elements of type WHAT. */

#define CP40_FREE_MAYBE(ptr) if (ptr) free(ptr)
/*	Free PTR only if PTR!=NULL. */

#define CP40_ALLOC_AT(ptr, number) ((ptr)= malloc((number)*sizeof((ptr)[0])))
/*	Allocate memory at PTR for NUMBER elements of same type as *PTR. */


#define CP40_BZERO(ptr, number) memset(ptr, 0, (number)*sizeof((ptr)[0]))
/*	Fills an array with zeroes (only works if PTR as the correct data
	type). */

/*---------------------------------------------------------------------------*/

/* Use definitions in "limits.h" to figure out a suitable data type for
   16/32-bit integers. */

#include <limits.h>

#define CP40_INT8_MIN                  (-128)
#define CP40_INT8_MAX                    127
#define CP40_UINT8_MAX                   255

#define CP40_INT16_MIN               (-32768L)
#define CP40_INT16_MAX                 32767L
#define CP40_UINT16_MAX                65535UL

#define CP40_INT32_MIN          (-2147483648L)
#define CP40_INT32_MAX            2147483647L
#define CP40_UINT32_MAX           4294967295UL

#undef CP40_INT16
#undef CP40_INT32

#if defined(SHRT_MIN) && defined(SHRT_MAX)
# if (SHRT_MIN == CP40_INT16_MIN) && (SHRT_MAX == CP40_INT16_MAX)
#  undef CP40_INT16
#  define CP40_INT16 short
# elif (SHRT_MIN == CP40_INT32_MIN) && (SHRT_MAX == CP40_INT32_MAX)
#  undef CP40_INT32
#  define CP40_INT32 short
# endif
#endif

#if defined(LONG_MIN) && defined(LONG_MAX)
# if (LONG_MIN == CP40_INT16_MIN) && (LONG_MAX == CP40_INT16_MAX)
#  undef CP40_INT16
#  define CP40_INT16 long
# elif (LONG_MIN == CP40_INT32_MIN) && (LONG_MAX == CP40_INT32_MAX)
#  undef CP40_INT32
#  define CP40_INT32 long
# endif
#endif

#if defined(INT_MIN) && defined(INT_MAX)
# if (INT_MIN == CP40_INT16_MIN) && (INT_MAX == CP40_INT16_MAX)
#  undef CP40_INT16
#  define CP40_INT16 int
# elif (INT_MIN == CP40_INT32_MIN) && (INT_MAX == CP40_INT32_MAX)
#  undef CP40_INT32
#  define CP40_INT32 int
# endif
#endif

#ifndef CP40_INT16
# error "cannot guess data type for 16-bit integer"
#endif

#ifndef CP40_INT32
# error "cannot guess data type for 16-bit integer"
#endif

/*---------------------------------------------------------------------------*/

_CP40_BEGIN_DECLS

/*---------------------------------------------------------------------------*/
/* PREDEFINED DATA TYPES */

typedef          CP40_INT16 cp40_int16_t;  /* 16 bit signed integer */
typedef unsigned CP40_INT16 cp40_uint16_t; /* 16 bit unsigned integer */
typedef          CP40_INT32 cp40_int32_t;  /* 32 bit signed integer */
typedef unsigned CP40_INT32 cp40_uint32_t; /* 32 bit unsigned integer */

typedef cp40_uint32_t             cp40_raw_photon_t;
typedef struct _cp40_raw_frame_t  cp40_raw_frame_t;
typedef struct _cp40_stream_t     cp40_stream_t;
typedef struct _cp40_stream_ops_t cp40_stream_ops_t;
typedef struct _cp40_io_file_t    cp40_io_file_t;
typedef struct _cp40_io_socket_t  cp40_io_socket_t;
typedef struct _cp40_instance_t   cp40_instance_t;
typedef struct _cp40_range_t      cp40_range_t;
typedef struct _cp40_window_t     cp40_window_t;
typedef struct _cp40_attr_entry_t cp40_attr_entry_t;
typedef struct _cp40_avg_stat_t   cp40_avg_stat_t;


/* A structure cp40_task_t is used to store all resources needed to perform
   a given photon-counting integration (e.g., long exposure, intercorrelation,
   autocorrelation, ..).  Each task is effectively performed onto the current
   pre-processed data by a routine with prototype cp40_task_op_t. */
typedef struct _cp40_task_t cp40_task_t;
typedef int cp40_task_op_t(cp40_task_t *task);


/* A structure cp40_stat_t is used to store raw-photons and statistics
   for input frames. */
typedef struct _cp40_stat_t cp40_stat_t;


/* Types defined for hash tables. */
typedef struct _cp40_hash_table  cp40_hash_table_t;
typedef struct _cp40_hash_entry  cp40_hash_entry_t;
typedef struct _cp40_hash_search cp40_hash_search_t;
typedef struct _cp40_hash_ops    cp40_hash_ops_t;

/*---------------------------------------------------------------------------*/
/* ERROR MANAGEMENT */

/* CP40 library store a single (the last one if any) error message in the
   private buffer `_cp40_error_message'.  When one of the library routines
   reports an error (e.g.  returns with a result different from CP40_OK)
   the error message can be obtained by `cp40_error_info(0/1)'.  The return
   value is a string pointer with last error message or NULL (no errors or
   last message cleared).  If the argument of `cp40_error_info' is
   non-zero, the error message is cleared as well (i.e.  future calls to
   `cp40_error_info' will return NULL unless a new error message is
   registered). */

/* Standard codes returned by CP40 routines: */
#define CP40_OK     0 /* Routine completes successfully. */
#define CP40_ERROR -1 /* Routine fails, generally, an error message is left
			that can be retrieved with cp40_error_info(CLEAR)
			-- the CLEAR argument indicates whether (CLEAR is
			non-zero) or not (CLEAR is zero) the error message
			should be discarded. */

#define CP40_ERROR_MESSAGE_LENGTH 1023 /* Maximum length of CP40 error
					  message. */
extern char _cp40_error_message[];     /* Private message buffer of size
					  CP40_ERROR_MESSAGE_LENGTH+1. */

extern int cp40_error(const char *format, ...)
     __attribute__((format (printf, 1, 2)));
/*	Store error message in private static buffer and return CP40_ERROR.
	The syntax is similar to that of fprintf(3).  BUG: the private
	error message buffer is large enough for small (useful) messages
	but there is no checking of writing a too long message... */

const char *cp40_error_info(int clear_error);
/*	Return last CP40 error message (the returned message is stored in a
	private static area and is only relevant just after an error occurs
	in CP40 library) or NULL if there is no error message.  If
	CLEAR_ERROR is non-zero the error status is cleared (i.e. subsquent
	calls will return NULL until a new error message is stored. */

extern int cp40_sys_error(const char *format, ...)
     __attribute__((format (printf, 1, 2)));
/*	Store formatted error message (FORMAT can be NULL, in which it is
	ignored) with a system (POSIX) error message appended.  The system
	error message is based on current value of global variable
	errno. */

extern void cp40_warning(const char *format, ...)
     __attribute__((format (printf, 1, 2)));
/*	Display warning message on standard error output (stderr).  The
	syntax is similar to that of fprintf(3). */

extern void cp40_fatal_error(const char *format, ...)
     __attribute__((format (printf, 1, 2)))
     __attribute__((noreturn));
/*	Display formatted error message on standard error output (stderr)
	and call abort().  The syntax is similar to that of fprintf(3). */

extern void cp40_assert_fail(const char *expr, const char *file, int line);
/*	Display message: "*** Assertion `EXPR' failed in FILE (LINE)."  on
	standard error output (stderr) and call abort(). */

extern const char *cp40_error_id(int err);
/*	Return a machine-readable textual identifier that corresponds to
	the given errno value ERR (e.g. "EPERM").  The identifier is the
	same as the #define name in `errno.h'. */

extern const char *cp40_error_msg(int err);
/*	Return a human-readable standard POSIX error message corresponding
	to the given errno value ERR.  This procedure is used instead of
	`strerror' because `strerror' returns slightly different values on
	different machines (e.g.  different capitalizations), which cause
	problems for things such as regression tests.  This procedure
	provides messages for most standard errors, then it calls
	`strerror' for things it doesn't understand. */

/*---------------------------------------------------------------------------*/
/* MEMORY MANAGEMENT */

extern void *cp40_alloc(unsigned int size, const char *what);
extern void *cp40_alloc_clear(unsigned int size, const char *what);
/* Allocate and return address of SIZE bytes dynamic memory.  In case of
   error, return NULL and left an error message that can be recovered with
   cp40_error_info.  If not NULL, WHAT is used to produce an the error message
   in the form: "insufficient memory for WHAT".  In addition, cp40_alloc_clear
   fill the allocated memory with zeroes. */

extern void *cp40_mem_alloc(size_t size);
/*	Allocate SIZE bytes of memory with malloc() and return address of
	allocated area or NULL on error.  In case of error,
	cp40_sys_error() is called to left an error message that can be
	retrieved by cp40_error_info(). */

extern void *cp40_mem_alloc_clear(size_t size);
/*	Same as cp40_mem_alloc(SIZE) but allocated memory is filled with
	zeroes. */

extern void *cp40_mem_realloc(void *ptr, size_t new_size);
/*	Allocate with malloc(), if PTR==NULL, or reallocate with realloc(),
	if PTR!=NULL, NEW_SIZE bytes of memory and return, possibly new,
	address of allocated area or NULL on error.  In case of error,
	memory previously allocated at PTR should be left unchanged(?)  and
	cp40_sys_error() is called to left an error message that can be
	retrieved by cp40_error_info(). */

extern void cp40_mem_free(void *ptr);
/*	Free PTR with free() only if non-NULL.  PTR must have been
	allocated with malloc(), cp40_alloc(), cp40_alloc_clear(), or
	cp40_mem_alloc() or cp40_mem_realloc() or one of the
	CP40_ALLOC/CP40_REALLOC macros). */

/*---------------------------------------------------------------------------*/
/* INDEXED 2D ARRAYS */

extern int cp40_alloc_2d(void *index_ptr, void *data, size_t elsize,
			 int imin, int imax, int jmin, int jmax);
/*	The function cp40_alloc_2d() allocates and instanciates an array of
	pointers for addressing 2D array at base address DATA.  If DATA is
	NULL, it is allocated and filled with zeroes by the function.
        A standard result, CP40_OK or CP40_ERROR is returned.  Assuming
	INDEX_PTR = &INDEX, then
	    INDEX[I][J]
	is a faster indirection than
	    DATA[(J - JMIN) + (JMAX - JMIN + 1)*(I - IMIN)]
	for IMIN <= I <= IMAX and JMIN <= J <= JMAX. */

extern void cp40_free_2d(void *index, int free_data, size_t elsize,
			 int imin, int imax, int jmin, int jmax);
/*	The function cp40_free_2d() free memory assoiciated with a 2D array
	allocated by cp40_alloc_2d().  If FREE_DATA is true, then the data
	memory is also released; otherwise only the memory allocated for
	the index get released. */


/*---------------------------------------------------------------------------*/
/* SUPPORT FOR XDR (eXternal Data Representation) */

/* Definitions of primary XDR data types. */
typedef unsigned char	cp40_xdr_byte_t;
typedef unsigned char	cp40_xdr_uchar_t;
typedef	         char	cp40_xdr_char_t;
typedef	cp40_int16_t	cp40_xdr_short_t;
typedef	cp40_uint16_t	cp40_xdr_ushort_t;
typedef	cp40_int32_t	cp40_xdr_int_t;
typedef	cp40_uint32_t	cp40_xdr_uint_t;
typedef	cp40_int32_t	cp40_xdr_long_t;
typedef	cp40_uint32_t	cp40_xdr_ulong_t;
typedef	float		cp40_xdr_float_t;
typedef	double		cp40_xdr_double_t;

/* Sizes (in bytes) of primary XDR data types. */
#define CP40_XDR_SIZEOF_BYTE   1
#define CP40_XDR_SIZEOF_CHAR   1
#define CP40_XDR_SIZEOF_UCHAR  CP40_XDR_SIZEOF_CHAR
#define CP40_XDR_SIZEOF_SHORT  2
#define CP40_XDR_SIZEOF_USHORT CP40_XDR_SIZEOF_SHORT
#define CP40_XDR_SIZEOF_INT    4
#define CP40_XDR_SIZEOF_UINT   CP40_XDR_SIZEOF_INT
#define CP40_XDR_SIZEOF_LONG   4
#define CP40_XDR_SIZEOF_ULONG  CP40_XDR_SIZEOF_LONG
#define CP40_XDR_SIZEOF_FLOAT  4
#define CP40_XDR_SIZEOF_DOUBLE 8

/* Routines to read/write XDR data. Behave like fread/fwrite but also take
   care of byte swapping data.  PTR must point to, at least, NUMBER elements
   of corresponding type. */
extern size_t cp40_xdr_read_byte(void *ptr, size_t number,
				 FILE *stream);
extern size_t cp40_xdr_write_byte(const void *ptr, size_t number,
				  FILE *stream);
extern size_t cp40_xdr_read_short(void *ptr, size_t number,
				  FILE *stream);
extern size_t cp40_xdr_write_short(const void *ptr, size_t number,
				   FILE *stream);
extern size_t cp40_xdr_read_long(void *ptr, size_t number,
				 FILE *stream);
extern size_t cp40_xdr_write_long(const void *ptr, size_t number,
				  FILE *stream);
extern size_t cp40_xdr_read_float(void *ptr, size_t number,
				  FILE *stream);
extern size_t cp40_xdr_write_float(const void *ptr, size_t number,
				   FILE *stream);
extern size_t cp40_xdr_read_double(void *ptr, size_t number,
				   FILE *stream);
extern size_t cp40_xdr_write_double(const void *ptr, size_t number,
				    FILE *stream);

/* Direct call for bytes. */
#define cp40_xdr_read_byte(ptr, number, stream)  fread(ptr, 1, number, stream)
#define cp40_xdr_write_byte(ptr, number, stream) fwrite(ptr, 1, number, stream)

/* In XDR, int's and long's are identical as well as char and byte: */
#define cp40_xdr_read_int   cp40_xdr_read_long
#define cp40_xdr_write_int  cp40_xdr_write_long
#define cp40_xdr_read_char  cp40_xdr_read_byte
#define cp40_xdr_write_char cp40_xdr_write_byte

/*---------------------------------------------------------------------------*/
/* BYTE ORDER AND BYTE SWAPPING */

extern int cp40_get_byte_order(void);
#define CP40_LITTLE_ENDIAN 1234
#define CP40_BIG_ENDIAN    4321
#define CP40_PDP_ENDIAN    3412

extern int cp40_swap_bytes(void *dst, const void *src,
			   size_t elsize, size_t number);
/*	Copy NUMBER elements of size ELSIZE from SRC to DST with byte
	swapping.  SRC and DST may be identical to perform in-place
	operation.  A standard result is returned (the routine only fails
	if ELSIZE is none of: 1, 2, 4, or 8 bytes). */

/*---------------------------------------------------------------------------*/
/* DYNAMIC BUFFERS */

/* IMPORTANT: Client users should always consider that the `cp40_dbuf_t'
              structure is opaque and should always use the provided
              rouines/macros to access/modify members of `cp40_dbuf_t'
              structure.  Also beware that cp40_dbuf_resize and
              cp40_dbuf_append may change the base address of dynamic
              buffer. */
typedef struct _cp40_dbuf cp40_dbuf_t;

#define CP40_DBUF_BASE(dbuf)      ((dbuf)->base)
/*	Get base address of data in dynamic buffer DBUF. */

#define CP40_DBUF_TAIL(dbuf)      ((void*)((char*)((dbuf)->base)+(dbuf)->size))
/*	Get starting address of free part of dynamic buffer. */

#define CP40_DBUF_AVAILABLE(dbuf) ((dbuf)->max_size - (dbuf)->size)
/*	Get number of unused bytes (i.e., size of `tail'). */

#define CP40_DBUF_SIZE(dbuf)      ((dbuf)->size)
/*	Get actual number of bytes stored in dynamic buffer DBUF. */

#define CP40_DBUF_MAX_SIZE(dbuf)  ((dbuf)->max_size)
/*	Get maximum storage size in dynamic buffer DBUF. */

#define CP40_DBUF_RESET(dbuf)     ((dbuf)->size= 0)
/*	Reset dynamic buffer DBUF to contain nothing. */

extern void cp40_dbuf_init(cp40_dbuf_t *dbuf);
/*	Initialize dynamic buffer DBUF (must be the first thing done).
	Contents of structure DBUF are arbitrarily all zero'ed. */

extern void cp40_dbuf_free(cp40_dbuf_t *dbuf);
/*	Release dynamic memory in dynamic buffer DBUF.  Dynamic memory, if
	any, gets freed; otherwise perform similarly to cp40_dbuf_init
	(i.e. the buffer is ready for future reuse). */

extern int cp40_dbuf_resize(cp40_dbuf_t *dbuf, size_t new_max_size);
/*	Make sure dynamic buffer is large enough to contain NEW_MAX_SIZE
	bytes.  Return standard result: CP40_OK or CP40_ERROR -- error
	message can be retrieved by cp40_error_info. It is an error to
	reduce maximum size of buffer below its actual count given by
	CP40_DBUF_SIZE(DBUF). */

extern int cp40_dbuf_append(cp40_dbuf_t *dbuf, void *data, size_t nbytes);
/*	Append NBYTES from buffer DATA into dynamic buffer DBUF.  The
	buffer is possibly resized.  Standard result is returned: CP40_OK
	or CP40_ERROR (error message can be retrieved with
	cp40_error_info). */

extern void cp40_dbuf_clear(cp40_dbuf_t *dbuf);
/*	Clear (set to 0) contents of DBUF. */

/* The true structure. */
struct _cp40_dbuf {
  size_t   size;     /* actual size of stored data (in bytes) */
  size_t   max_size; /* maximum storage available (in bytes) */
  void    *base;     /* base address of data stored in dynamic buffer */
};

#define CP40_DBUF_GRAIN 128 /* When a dynamic buffer needs to be augmented,
			      it is resized by a multiple of this number of
			      bytes. */

/*---------------------------------------------------------------------------*/
/* CHARACTER ROUTINES */

extern int cp40_to_lower(int c);
extern int cp40_to_upper(int c);
/*----- These functions convert their argument to lower and upper case
	letter respectively (except that EOF (-1) remains EOF).  Assume
	ISO-Latin1 encoding for accentuated letters.  Also implemented as
	macros/inline functions.  Use CP40_TO_LOWER/CP40_TO_UPPER if you want
	to be sure to use the macro verison. */

extern int cp40_is_space(int c);
/*----- Check if character C is a space (any of '\t', '\n', '\v', '\f',
	'\r' or an ordinary space) or not.  Also implemented as
	macros/inline functions.  Use CP40_IS_SPACE if you want to be sure
	to use the macro verison.  Note: cp40_is_space(0) and
	CP40_IS_SPACE(0) are both false (as expected ;-). */

extern const short *_cp40_to_lower, *_cp40_to_upper;
extern const char  *_cp40_is_space;
/*----- These (private although globally accessible) pointers are used for
	fast character check/conversion and can be indexed in the range
	-128..255; so that their index -- the argument of the corresponding
	character macros/function -- can be either a signed or unsigned
	char.  Internally, static arrays of 384 elements are used. */

/* Character macros. */
#define CP40_IS_SPACE(c) (_cp40_is_space[(int)(c)])
#define CP40_TO_LOWER(c) (_cp40_to_lower[(int)(c)])
#define CP40_TO_UPPER(c) (_cp40_to_upper[(int)(c)])

/*---------------------------------------------------------------------------*/
/* STRING ROUTINES */

extern int cp40_str_get_integer(const char *string, long *long_ptr);
extern int cp40_str_get_real(const char *string, double *double_ptr);
/*----- Convert zero terminated input string STRING into a long integer (a
	double precision real) stored at LONG_PTR (resp. DOUBLE_PTR).  The
	string may contains leading/trailing spaces.  If the conversion
	succeeds, CP40_OK is returned; otherwise, CP40_ERROR is returned and
	the error message can be retrieved by cp40_error_info(). */

extern char *cp40_str_clone(const char *string);
/*----- Return dynamically allocated memory with a copy of STRING.  NULL is
	returned in case of error or if STRING is NULL. */

extern char *cp40_str_print_int(char *string, int value);
extern char *cp40_str_print_long(char *string, long value);
/*----- Print into STRING base 10 human readable representation of int/long
	VALUE.  About 3 times faster than `sprintf("%d", STRING,
	VALUE)'. */

extern char *cp40_str_upper(register char *dst, register const char *src);
extern char *cp40_str_lower(register char *dst, register const char *src);
/*----- Convert zero terminated input string SRC to lower/upper case
	letters (NULL string are the same as empty ones) and store the
	result in DST.  Return DST.  Operation can be done `in-place': with
	DST==SRC. */

extern char *cp40_str_trim(char *dst, const char *src, int flags);
#define CP40_STR_TRIM_LEFT  1
#define CP40_STR_TRIM_RIGHT 2
#define CP40_STR_TRIM_BOTH  (CP40_STR_TRIM_LEFT|CP40_STR_TRIM_RIGHT)
/*----- Copy zero-terminating string SRC into DST removing leading (bit
	CP40_STR_TRIM_LEFT of FLAGS is set) and/or trailing (bit
	CP40_STR_TRIM_RIGHT of FLAGS is set) spaces.  Returns DST. */

extern char *cp40_str_truncate(char *buf, const char *string, int maxlen,
			      const char *ellipsis);
/*----- Copy string with possible truncation.  If strlen(STRING) <= MAXLEN,
	copy STRING into BUF; otherwise, copy MAXLEN-strlen(ELLIPSIS)
	characters of STRING into BUF and append ELLIPSIS (or "[...]"  if
	ELLIPSIS is NULL).  Return BUF or NULL if STRING is NULL or MAXLEN
	<= strlen(ELLIPSIS) (or 5 if ELLIPSIS is NULL).  BUF must be at
	least MAXLEN+1 byte long.  */

extern const char *cp40_str_ordinal_suffix(int number);
/*----- Return static string "st" or "nd" or "rd" or "th". */

extern char *cp40_str_token(char *string, char **next, int *sval,
			 const char *slist);
/*----- Return next token from STRING or NULL if token list is exhausted.
	Tokens are separated by characters found in SLIST or spaces if
	SLIST is NULL (see CP40_IS_SPACE for the list of space characters).
	The end of the returned token is marked by a '\0' in STRING, if
	SVAL is not NULL the value of the separator that terminates the
	token is stored in *SVAL.  NEXT the address of a string pointer
	used to keep track of the position in STRING where to start
	searching next token: at first call *NEXT must be NULL and *NEXT
	must not be modified between calls to cp40_str_token. */

/*---------------------------------------------------------------------------*/
/* DYNAMIC STRING ROUTINES */

/* Structure to store a dynamic string.  Clients are not allowed to directly
   access, members of this structure (use the below defined macros instead). */
typedef struct _cp40_dstr cp40_dstr_t;
struct _cp40_dstr {
    char *string; /* Pointer to cp40_alloc'ed memory or to the internal static
		     buffer. */
    int   length; /* Actual string length. */
    int   maxlen; /* Maximum string length (it is guaranted that there is an
		     extra character left for the terminating zero). */
#define CP40_DSTR_STATIC_LENGTH 127
    char  buffer[CP40_DSTR_STATIC_LENGTH+1]; /* Internal static buffer. */
};

/* Clients should only use the following macros to access dynamic string
   members (DSP is the address of a valid -- initialized -- dynamic string
   structure). */
#define cp40_dstr_value(dsp)	((dsp)->string)
#define cp40_dstr_length(dsp)	((dsp)->length)
#define cp40_dstr_maxlen(dsp)	((dsp)->maxlen)

extern void cp40_dstr_init(cp40_dstr_t *dsp);
/*----- Initializes a dynamic string, discarding any previous contents of
	the string (cp40_dstr_free should have been called already if the
	dynamic string was previously in use). */

extern void cp40_dstr_free(cp40_dstr_t *dsp);
/*----- Frees up any memory allocated for the dynamic string and
	reinitializes the string to an empty state pointing to the internal
	buffer. */

extern cp40_dstr_t *cp40_dstr_reset(cp40_dstr_t *dsp);
/*----- Reset dynamic string to be empty (i.e. of zero length) and return
	its argument. */

extern int cp40_dstr_resize(cp40_dstr_t *dsp, int length);
/*----- Resizes dynamic string buffer so that it can store at least strings
	of LENGTH characters.  Returns standard result: CP40_OK or CP40_ERROR
	-- in that case, the error message can be retrieved with
	cp40_error_info(). */

extern int cp40_dstr_append_char(cp40_dstr_t *dsp, int c);
/*----- Append a single character C to the dynamic string DSP.  Returns
	standard result: CP40_OK or CP40_ERROR -- in that case, the error
	message can be retrieved with cp40_error_info(). */

extern int cp40_dstr_append(cp40_dstr_t *dsp, const char *string, int length);
/*----- Appends STRING to the dynamic string DSP.  If LENGTH is less than
	zero, strlen(STRING) is used instead; otherwise only LENGTH
	characters (including NULL's if any) are copied to the end of DSP
	(plus a terminating zero).  Returns standard result: CP40_OK or
	CP40_ERROR -- in that case, the error message can be retrieved with
	cp40_error_info(). */

extern int cp40_dstr_append_int(cp40_dstr_t *dsp, int value, int separator);
extern int cp40_dstr_append_ints(cp40_dstr_t *dsp, const int array[],
				int number, int separator);
extern int cp40_dstr_append_double(cp40_dstr_t *dsp, double value,
				  int separator);
extern int cp40_dstr_append_doubles(cp40_dstr_t *dsp, const double array[],
				   int number, int separator);
/*----- Convert value(s) into human readable ASCII string taking care that
	floating-point values do not look like integer and append them to
	DSP.  If SEPARATOR is not '\0' this character is used as a
	separator before each value but the first one (i.e. the one which
	starts the resulting string).  Return standard result: CP40_OK or
	CP40_ERROR -- in that case, the error message can be retrieved with
	cp40_error_info(). */


extern char *cp40_dstr_steal(cp40_dstr_t *dsp);
/*----- Steal contents of dynamic string pointed by DSP.  Return a dynamic
	string that must be cp40_free'd by the caller and reset input
	dynamic string.  In case of error, returns NULL, and error message
	can be retrieved with cp40_error_info(). */

extern int cp40_dstr_read_line(cp40_dstr_t *dsp, FILE *stream, int *eof_ptr,
			      int *line_count_ptr);
/*----- Reads next text line from STREAM and stores it in dynamic string
	DSP.  In actual files and devices the end of a line may be
	represented differently on different platforms, or even for
	different devices on the same platform. For example, under UNIX
	newlines (\n) are used in files, whereas carriage-return-linefeed
	(\r\n) sequences are normally used in network connections.  This
	routine recognizes, in that order, EOF, \r\n, \r or \n sequences as
	valid end-of-line and replaces them by a null char (\0).

	If EOF_PTR is non-NULL, its value is set to 1 upon end-of-file and
	to 0 otherwise.  If LINE_COUNT_PTR is non-NULL, its value is
	incremented if a new line is actually read.

	Returns standard result: CP40_OK or CP40_ERROR -- in that case, left
	an error message that can be retrieved with cp40_error_info(). */

extern void cp40_dstr_lower(cp40_dstr_t *dsp);
extern void cp40_dstr_upper(cp40_dstr_t *dsp);
/*----- Converts dynamic string DSP to lower/upper case characters. */

extern void cp40_dstr_trim(cp40_dstr_t *dsp, int flags);
/*----- Removes leading (bit CP40_STR_TRIM_LEFT of FLAGS is set) and/or
	trailing (bit CP40_STR_TRIM_RIGHT of FLAGS is set) spaces from
	DSP. */

extern void cp40_dstr_pack(cp40_dstr_t *dsp);
/*----- Replace all occurences of multiple space characters in DSP by a
	single ordinary space. */

/*---------------------------------------------------------------------------*/

#if defined(HAVE_INLINE) && HAVE_INLINE
extern inline int cp40_to_lower(int c) { return (int)CP40_TO_LOWER(c); }
extern inline int cp40_to_upper(int c) { return (int)CP40_TO_UPPER(c); }
extern inline int cp40_is_space(int c) { return (int)CP40_IS_SPACE(c); }
#else /* not HAVE_INLINE */
# define cp40_to_lower(c) ((int)CP40_TO_LOWER(c))
# define cp40_to_upper(c) ((int)CP40_TO_UPPER(c))
# define cp40_is_space(c) ((int)CP40_IS_SPACE(c))
#endif /* not HAVE_INLINE */

/*---------------------------------------------------------------------------*/
/* FILE MANAGEMENT */

#define CP40_IDENT_SIZE 16
extern int cp40_write_file_version(FILE *file, const char *name, int major,
				   int minor);
extern int cp40_read_file_version(FILE *file, char *name, int *major,
				  int *minor);

/* In order to avoid having files that remains open (because the routine
   which opens them returns without closing them) a solution is to use a
   static pointer to a FILE structure (defined in `stdio.h') and
   initialized to NULL (this is IMPORTANT) and to use the following
   routines for open/close this file.

   For instance:
     (1) declaration of the static file pointer:
         static FILE *stf= NULL;
     (2) open this static file:
         if (cp40_open_static_file(&stf, filename, mode) != CP40_OK) {
	   fputs(cp40_error_info(1), stderr);
	   exit(1);
	 }
     (3) to close the static file, just do:
	 cp40_close_static_file(&stf); */

extern int cp40_open_static_file(FILE **file_ptr, const char *filename,
				 const char *mode);
/*	Open the file named FILENAME for the mode MODE (see fopen(3)) and
	store stream address in *FILE_PTR, keeping care of closing file
	stream (that's why it is important to initialize it with NULL) and
	return CP40_OK or CP40_ERROR (with error message stored). */

extern void cp40_close_static_file(FILE **file_ptr);
/*	Closes the static file at *FILE_PTR taking care if it is already
	closed (i.e., set to NULL). */

/*---------------------------------------------------------------------------*/
/* HASH TABLES */

#define CP40_HASH_STRING_KEYS     0
#define CP40_HASH_ONE_WORD_KEYS   1
extern void cp40_hash_init_table(cp40_hash_table_t *table, int key_type);
/*	Creates a new hash table and stores information about the table at
	*TABLE.  KEY_TYPE is either CP40_HASH_STRING_KEYS,
	*CP40_HASH_ONE_WORD_KEYS, or an integer greater than 1 */

extern void cp40_hash_delete_table(cp40_hash_table_t *table);
/*	Deletes all the entries in the hash table and frees up related
	storage. */

#define cp40_hash_create_entry(table, key, new) \
    (table)->ops->create(table, key, new)
/*	Returns a pointer to the entry in TABLE whose key is KEY, creating
	a new entry if needed.  *NEW is set to 1 if a new entry was created
	or 0 if the entry already existed. */

extern void cp40_hash_delete_entry(cp40_hash_entry_t *entry);
/*	Deletes ENTRY from its hash table.  ENTRY should never again be
	used by the caller. Call ENTRY->delete to free ENTRY->client_data
	if such a callback was registered, otherwise it is up to the caller
	to free the ENTRY->client_data member of the entry, if that is
	relevant. */

extern void cp40_hash_set_value(cp40_hash_entry_t *entry, void *client_data,
			 void (*delete)(void *client_data));
/*	Sets the value associated with a hash table entry as well as its
	'delete' method.  Which can be a user routine or one of: */
#define CP40_HASH_DYNAMIC cp40_mem_free	/* Delete if non-null. */
#define CP40_HASH_STATIC  NULL	        /* Do nothing. */

#define cp40_hash_get_value(h) ((h)->client_data)
/*	Returns the value associated with a hash table entry H. */

#define cp40_hash_get_key(h)   \
    ((char *) (((h)->table->key_type == CP40_HASH_ONE_WORD_KEYS) \
	    ? (h)->key.one_word_value : (h)->key.string))
/* Returns the key associated with a hash table entry H. */

#define cp40_hash_find_entry(table, key)  (table)->ops->find(table, key)
/*	Returns a pointer to the entry in TABLE whose key is KEY, or NULL
	if no such entry exists. */

extern cp40_hash_entry_t *cp40_hash_first_entry(cp40_hash_table_t *table,
						cp40_hash_search_t *search);
/*	Starts a search through all the elements of a hash table. Stores
	information about the search at *search_ptr and returns the hash
	table's first entry or NULL if it has no entries. */

extern cp40_hash_entry_t *cp40_hash_next_entry(cp40_hash_search_t *search);
/*	Returns the next entry in the search identified by SEARCH or NULL
	if all entries in the table have been returned. */


extern char *cp40_hash_stats(cp40_hash_table_t *table);
/*	Returns a string giving usage statistics for TABLE.  The string is
	dynamically allocated and must be freed by the caller. */

/* Structure definition for an entry in a hash table.  No-one outside the
   CP40 routines should access any of these fields directly; use the
   defined macros instead. */
struct _cp40_hash_entry {
  cp40_hash_entry_t *next_ptr;    /* Pointer to next entry in this hash
				     bucket, or NULL for end of chain. */
  cp40_hash_table_t *table;       /* Pointer to table containing entry. */
  cp40_hash_entry_t **bucket_ptr; /* Pointer to bucket that points to first
				     entry in this entry's chain: used for
				     deleting the entry. */
  void *client_data;              /* Application stores something here with
				     cp40_hash_set_value. */
  void (*delete)(void *);         /* Method called to free client_data. */
  union { /* Key has one of these forms: */
    const char *one_word_value;   /* One-word value for key. */
    int words[1];	          /* Multiple integer words for key.  The
				     actual size will be as large as
				     necessary for this table's keys. */
    char string[4];	          /* String for key.  The actual size will
				     be as large as needed to hold the
				     key. */
  } key;		          /* MUST BE LAST FIELD IN RECORD!! */
};

/* Structure definition for a hash table.  Must be in this header file so
   clients can allocate space for these structures, but clients should
   never access any fields in this structure.  */
struct _cp40_hash_table {
  cp40_hash_entry_t **buckets; /* Pointer to bucket array.  Each element
				  points to first entry in bucket's hash
				  chain, or NULL. */
#define CP40_HASH_SMALL_TABLE 4
  cp40_hash_entry_t *static_buckets[CP40_HASH_SMALL_TABLE];
  /* Bucket array used for small tables (to avoid mallocs and frees). */
  int num_buckets;	/* Total number of buckets allocated at **buckets. */
  int num_entries;	/* Total number of entries present in table. */
  int rebuild_size;	/* Enlarge table when num_entries gets to be this
			   large. */
  int down_shift;	/* Shift count used in hashing function.  Designed
			   to use high-order bits of randomized keys. */
  unsigned int mask;	/* Mask value used in hashing function. */
  int key_type;	        /* Type of keys used in this table.  It's either
			   CP40_HASH_STRING_KEYS, CP40_HASH_ONE_WORD_KEYS,
			   or an integer giving the number of ints that is
			   the size of the key. */
  const cp40_hash_ops_t *ops;  /* Hash operations table. */
};

/* Hash operations table. */
struct _cp40_hash_ops {
  cp40_hash_entry_t *(*find)(cp40_hash_table_t *table, const char *key);
  cp40_hash_entry_t *(*create)(cp40_hash_table_t *table, const char *key,
			       int *new);
};

/* Structure definition for information used to keep track of searches through
   hash tables: */
struct _cp40_hash_search {
  cp40_hash_table_t *table;      /* Table being searched. */
  int next_index;	         /* Index of next bucket to be enumerated
				    after present one. */
  cp40_hash_entry_t *next_entry; /* Next entry to be enumerated in the current
				    bucket. */
};

/*---------------------------------------------------------------------------*/
/*
 * SIMPLE EDITABLE DATABASE FILES
 * ==============================
 *
 * CP40 library implements a simple mechanism to save hash tables into
 * editable database files.  As far as the hash table contents is pure text
 * (both keys and values), the database files are editable by hand (i.e.,
 * with your favorite text editor) since they follow a very simple syntax:
 *   - A line that begins with a hash '#' (after any number of spaces) is
 *     considered as being a comment (treated as a blank line).
 *   - Hash entry key and value are specified by:
 *       KEY=VALUE
 *     where KEY is made of any characters (except '#', '>', '=' or space).
 *     There may be any number of spaces before and/or after KEY or the
 *     equal '=': VALUE starts with the first non-space character after the
 *     '=' and up to the end-of-line.  A too long value can be continued on
 *     the next line by starting it with a '>'.  If there are no non-space
 *     character after the '>', a newline '\n' is appended to VALUE;
 *     otherwise every character, including spaces, after the '>' is
 *     considered to be part of VALUE up to the end of line.  In order to
 *     keep track of previous settings, the value of the current entry can
 *     be reset by a line starting with a '=' as in:
 *       KEY = OLD_VALUE
 *           = NEW_VALUE
 *           = NEWER_VALUE
 *     There must be no blank/comment lines between successive new values
 *     for a card entry.
 *   - Blank lines are ignored.
 * In order to handle newlines in VALUE, the convention is to convert
 * newline into empty continuation line (and reciprocally).
 *
 * Another improvement with respect to actual implementation of simple hash
 * tables is that I tried to do the necessary housekeeping to avoid
 * problems if some of these routines get interrupted, e.g. by user typing
 * ^C.  This may seem too paranoid but is really needed if the CP40 library
 * is linked with some interactive language, e.g. Yorick or IDL, which
 * allow such interruptions.
 */

extern int cp40_hash_restore(cp40_hash_table_t *table, const char *filename);
/*	Reads editable database file and store its contents in hash table
	pointed by TABLE which is assumed to have been correctly
	initialized -- i.e., with cp40_hash_init_table(TABLE,
	CP40_HASH_STRING_KEYS) -- and of which the contents may be freed
	before overwriten -- i.e., all entries that already exit in the
	table must have a dynamically allocated string value (or NULL) that
	is referenced nowhere else.  Returns standard result: CP40_OK or
	CP40_ERROR.  Beware that the restored values are dynamically
	malloc'ed and that value of existing entries are ckfree'd before
	being overwritten. */

extern int cp40_hash_save(cp40_hash_table_t *tabl, const char *filename,
			  int overwrite);
/*	Saves contents of hash table pointed by TABLE to editable database
	file FILENAME and returns standard result: CP40_OK or CP40_ERROR.
	If OVERWRITE is zero, FILENAME must not exists.  This routine is a
	very simple driver, you may prefer using the following routines to
	add your own comments. */

extern int cp40_hash_write_entry(FILE *file, const char *key,
				 const char *value, int comment);
/*	Print out a hash entry key-value pair taking care of splitting
	value if it is too long (if COMMENT is non-zero the resulting line
	get commented).  String pointer NAME can be NULL (same as
	empty/blank string).  Returns standard result: CP40_OK or
	CP40_ERROR. */

extern int cp40_hash_write_comment(FILE *file, const char *string);
/*	Print out a comment line in a database file (STRING must not
	contain any newline; this is not checked).  Returns standard
	result: CP40_OK or CP40_ERROR. */

/*---------------------------------------------------------------------------*/
/* SOCKET CONNECTION */

/* These bits may be OR'ed together into the "flags" argument of the
   cp40_create_socket() function: */
#define CP40_SOCK_ASYNC  (1<<0) /* Asynchronous client-socket connection. */
#define CP40_SOCK_SERVER (1<<1) /* Create a server socket. */
#define CP40_SOCK_CLIENT (1<<2) /* Create a client socket. */

extern int cp40_create_socket(int port, const char *host, const char *myaddr,
			     int myport, int flags);
/*	This function opens a new socket in client or server mode and
	initializes it. Returns -1 if an error occurs; otherwise, returns
	a valid path descriptor referencing the socket. Arguments are:
	  PORT   Port number to open
	  HOST   Name of host on which to open port.  NULL implies INADDR_ANY.
	  MYADD  Optional client-side address.
	  MYPORT Optional client-side port.
	  FLAGS  OR'ed bit flags.  If CP40_SOCK_SERVER is set, create a server
	         socket; otherwise, create a client socket.  For a client
		 socket, if CP40_SOCK_ASYNC is set, attempt to do an
		 asynchronous connect; otherwise, do a synchronous connect
		 or bind. */

/*---------------------------------------------------------------------------*/
/* CP40 HARDWARE SETTINGS (CCD sizes, etc). */

#define CP40_NCHANNELS       4	/* Number of CP40 channels. */
#define CP40_CCD_WIDTH     384
#define CP40_CCD_HEIGHT	   288
#define CP40_WIDTH(pixel)  (CP40_CCD_WIDTH  << (2 - (pixel)))
#define CP40_HEIGHT(pixel) (CP40_CCD_HEIGHT << (2 - (pixel)))
/*	Return the number of raw pixels for one CP40 channel along the
	vertical/horizontal direction; the argument is one of the
	CP40_PIXEL_... values (see below). */

/*---------------------------------------------------------------------------*/
/* Window settings (if one of the bounds is equal to CP40_WINDOW_WILD, then
   the corresponding extreme value should be used instead). */
struct _cp40_window_t {
# define CP40_WINDOW_WILD (0x80000000L)
  int xmin, xmax, ymin, ymax;
};

/* Interval of values (the flags indicate if a particular bound is set).  */
struct _cp40_range_t {
# define CP40_MIN_SET  ((unsigned int)0x1)
# define CP40_MAX_SET  ((unsigned int)0x2)
  unsigned int flags;
  double min, max;
};

/* Constants for header sizes.  */
#define CP40_HEADER_SIZE     0x4000	/* 16 Kbytes header. */
#define CP40_TEXT_OFFSET     0x0400	/*  1 Kbytes binary header. */

/*---------------------------------------------------------------------------*/
			 /**************************
			  * PRE-PROCESSING OF DATA *
			  **************************/

/*---------------------------------------------------------------------------*/
/* SELECTION OF FRAME PARITY */

/* Frame selection based on the parity of the frame number (date). */
#define CP40_PARITY_EVEN  1
#define CP40_PARITY_ODD	  2
#define CP40_PARITY_ANY   (CP40_PARITY_EVEN|CP40_PARITY_ODD)


/* Routines to query/set parity of selected frames. */
extern int cp40_get_parity(cp40_instance_t *instance);
extern int cp40_set_parity(cp40_instance_t *instance, int parity);


/*---------------------------------------------------------------------------*/
/* RAW PIXEL RESOLUTION */

typedef enum {
  /* Raw-pixel  sampling (IMPORTANT:  The following values  are assumed  to be
     equal to the number of right  shift bits needed to convert data raw pixel
     coordinates,  i.e. logical,  to  a particular  sampling.   DO NOT  CHANGE
     THEM). */
  CP40_PIXEL_LOGICAL=       0,	/* Raw pixel resolution given by the CP40
				   camera after centroiding (MUST BE 0). */
  CP40_PIXEL_HALF_PHYSICAL= 1,	/* 2 logical pixels (MUST BE 1). */
  CP40_PIXEL_PHYSICAL=      2	/* 4 logical pixels, also the CCD resolution
				   of the CP40 (*MUST* be 2). */
} cp40_pixel_t;

/* Routines to query/set raw-pixel sampling. */
extern cp40_pixel_t cp40_get_pixel(cp40_instance_t *instance);
extern int cp40_set_pixel(cp40_instance_t *instance, cp40_pixel_t value);

/*---------------------------------------------------------------------------*/
/* SPURIOUS PHOTONS */

/* In principle,  the centroiding  device  forbid  the  detection of  multiple
   events in  the same  frame within  the same small  region (3x3  CCD pixel).
   This produce a bias known  as the "trou du centreur" (photon-counting hole)
   in autocorrelations of  frames.  The first step of  the data pre-processing
   consist in dealing with such "spurious" events. */

typedef enum {
  /* Possible methods to deal with spurious photo-events: */
  CP40_SPURIOUS_LEAVE=   1, /* Do nothing special about spurious (i.e.
			       assume there is no spurious events). */
  CP40_SPURIOUS_DELETE=  2, /* Delete all spurious photo-events. */
  CP40_SPURIOUS_CENTRE=  3, /* Replace group of spurious events by a single
			       one with average position. */
  CP40_SPURIOUS_SINGLE=  4  /* Replace group of spurious events by first
			       of them. */
} cp40_spurious_t;

/* Routines to query/set method which deals with spurious photo-events. */
extern cp40_spurious_t cp40_get_spurious(cp40_instance_t *instance);
extern int cp40_set_spurious(cp40_instance_t *instance, cp40_spurious_t value);

/*---------------------------------------------------------------------------*/
/* DISTORTION CORRECTION */

/* The image intensifier (mainly the 1st generation tube) of the CP40 is
   responsible for a geometrical distortion that can be fixed thanks to a 2-D
   polynomial correction.  A 2rd order correction is sufficient:

     X' = CX[0] + CX[1]*X + CX[2]*Y + CX[3]*X*X + CX[4]*X*Y + CX[5]*Y*Y
        + CX[6]*X*X*X +  CX[7]*X*X*Y + CX[8]*X*Y*Y + CX[9]*Y*Y*Y
     Y' = CY[0] + CY[1]*X + CY[2]*Y + CY[3]*X*X + CY[4]*X*Y + CY[5]*Y*Y
        + CY[6]*X*X*X +  CY[7]*X*X*Y + CY[8]*X*Y*Y + CY[9]*Y*Y*Y

   For  efficiency  factorized  expressions   should  be  used  instead.   The
   following macros give  the coordinates (X',Y') of the  pixel corrected from
   the distortion,  C is  the array of  coefficients of  distortion correction
   (double C[2*CP40_NCOEFS], i.e., CY=CX+CP40_NCOEFS), `X' and `Y' are the raw
   *PHYSICAL*  pixel  coordinates (eventually  floating  values for  sub-pixel
   position).  */
#define CP40_XCOR(c,x,y) \
    ((c)[0]+((c)[1]+(c)[4]*(y)+((c)[3]+(c)[6]*(x)+(c)[7]*(y))*(x))*(x)\
     +((c)[2]+((c)[5]+(c)[8]*(x)+(c)[9]*(y))*(y))*(y))
#define CP40_YCOR(c,x,y) \
    ((c)[10]+((c)[11]+(c)[14]*(y)+((c)[13]+(c)[16]*(x)+(c)[17]*(y))*(x))*(x)\
     +((c)[12]+((c)[15]+(c)[18]*(x)+(c)[19]*(y))*(y))*(y))

#define CP40_NCOEFS      10 /* number of distortion correction coefficients
			       per polynomial */
#define CP40_COEF_COUNT (2*CP40_NCHANNELS*CP40_NCOEFS)
#define CP40_COEF_SIZE  (sizeof(double) * CP40_COEF_COUNT)

typedef enum {
  /* Possible distortion correction methods: */
  CP40_CORRECTION_NONE=     1, /* No distortion correction, the pixel sampling
				  is defined by the "pixel" attribute. */
  CP40_CORRECTION_SIMPLE=   2, /* Transformed positions get rounded to pixel
				  precision. */
  CP40_CORRECTION_RANDOM=   3, /* A sub-raw-pixel position is randomly
				  generated, the transformed positions get
				  rounded to pixel precision. */
  CP40_CORRECTION_WEIGHTED= 4  /* Each photon gives several positions with
				  different weights, those weights and
				  positions are tabulated. */
} cp40_correction_t;

/* Routines to query/set method used for distortion correction. */
extern cp40_correction_t cp40_get_correction(cp40_instance_t *instance);
extern int cp40_set_correction(cp40_instance_t *instance,
			       cp40_correction_t value);

extern double cp40_get_angle(cp40_instance_t *instance);
extern int    cp40_set_angle(cp40_instance_t *instance, double angle);
/* Query or set rotation angle (in degrees anti-clockwise) by
   distortion correction.  The "set" routine return standard result:
   CP40_OK or CP40_ERROR. */

extern double cp40_get_scale(cp40_instance_t *instance);
extern int    cp40_set_scale(cp40_instance_t *instance, double scale);
/* Query or set magnification scale (in pixels per grid step) by distortion
   correction.  The "set" routine return standard result: CP40_OK or CP40_ERROR.
   A value SCALE=0 means no distortion correction. */

extern int cp40_set_raw_coefs(cp40_instance_t *instance, double *raw_coefs);
extern void cp40_get_raw_coefs(cp40_instance_t *instance, double *raw_coefs);
extern void cp40_get_coefs(cp40_instance_t *instance, double *coefs);
/* Query/set raw (unscaled) or effective (scaled) distortion correction
   coefficients.  There are CP40_COEF_COUNT coefficients (CP40_NCOEFS per
   axis per channel). */

extern int cp40_shift(cp40_instance_t *instance, int channel,
		      double dx, double dy);
/* Modify raw coefficients of INSTANCE so that undistorted image of CHANNEL
   gets shifted by offset (DX,DY).  The offsets are given in undistorted
   pixel units (i.e.  the routine takes care of the current ANGLE and SCALE
   attributes of INSTANCE). */

extern int cp40_reweight_table(cp40_instance_t *this, int channel,
			       const double weight[],
			       int width, int height);
/* Apply flat field correction to weighted distortion correction.  WEIGHT
   is a WIDTH by HEIGHT array of weights (= 1/flat) computed _without_
   distortion correction. The dimensions must match those of the CCD
   for the raw pixel resolution as set in THIS. */

/*---------------------------------------------------------------------------*/
/* PRE-PROCESSED PHOTON FRAME */

typedef struct _cp40_frame_t   cp40_frame_t;
typedef struct _cp40_photon_t  cp40_photon_t;

#define CP40_FORCE_REPROCESSING 0x8
extern cp40_frame_t *cp40_get_frame(cp40_instance_t *instance, int flags);
/* Return a frame of preprocessed events from current raw (super-)frame of
   INSTANCE.  FLAGS is the channel number, possibly OR'ed with
   CP40_FORCE_REPROCESSING (to skip setp 1 below). The following steps are
   involved:
     1. If INSTANCE already has a frame for CHANNEL synchronized with
        current raw-frame date, the address of this frame is returned.
     2. INSTANCE->frame[CHANNEL] is created if it does not exists.
     3. Selection based on "parity": if INSTANCE->parity does not match the
        parity of INSTANCE->frame_date, INSTANCE->frame[CHANNEL] is
        returned with INSTANCE->frame[CHANNEL]->status=0 (frame was not
        selected) and INSTANCE->frame[CHANNEL]->count=0 (no events).
     4. Select raw-events (cp40_raw_photon_t) from CHANNEL and convert
        them to event (cp40_photon_t), deal with "spurious" events and
        raw-pixel "pixel" resolution.
     5. Apply distortion "correction", and "window" rules.
     6. Compute frame statistics.
     7. Apply other selection rules ("flux", "spread", "distance", "ratio").

   Return value is INSTANCE->frame[CHANNEL] (see FRAME->status for different
   selection cases) or NULL on error (error message can be recovered by
   cp40_error_info). */

struct _cp40_frame_t {
  /* Settings for photons selection: after a call to cp40_preprocess,
     FRAME->buffer is filled with photons from FRAME->channel and with
     positions in [FRAME->xmin:FRAME->xmax, FRAME->ymin:FRAME->ymax]. */

  cp40_instance_t *owner;     /* where events come from */
  int channel;                /* selected channel */
  int xmin, xmax, ymin, ymax; /* coordinate boundaries */
  int status;                 /* after a call to cp40_preprocess():
                                  0 if the current frame was rejected (check
                                    FRAME->count to see if the selection occurs
				    before/after the preprocessing of events);
				  1 if the, possibly empty, current frame was
				    effectively selected;
				 -1 on error */
  cp40_photon_t *buffer;      /* buffer of selected and preprocessed photons */
  int size;	              /* buffer size (maximum count) */
  int date;                   /* frame number */
  int count;                  /* actual number of photons in buffer */

  /* Frame statistics. */
  double flux;	              /* total of photon weights (set to -1 when
				 the frame has not yet been pre-processed). */
  double x, y;                /* photo-centre coordinates */
  double spread;              /* spread of photons (in pixels) */
  double distance;            /* distance of photo-centre to the closest
				 edge */
  double ratio;               /* ratio FRAME->distance / FRAME->spread */
};

struct _cp40_photon_t {
  /* Photon-event as stored in cp40_frame_t. */
  double w;	/* weight */
  int    x;	/* abscissa */
  int    y;	/* ordinate */
};


/*---------------------------------------------------------------------------*/
/* DATA STREAM */

extern void cp40_init_stream(cp40_stream_t *stream);
/*	Initialize CP40 stream access informations.  Must be called before
	using the stream. */

extern int cp40_assoc_file(cp40_instance_t *instance, const char *name,
			   int mode);
extern int cp40_assoc_socket(cp40_instance_t *instance, const char *host,
			      int port);
extern int cp40_close_stream(cp40_instance_t *instance, int update);

extern int cp40_read_header(cp40_instance_t *instance);
extern int cp40_write_header(cp40_instance_t *instance);
extern int cp40_set_exposure(cp40_instance_t *instance, int exposure);

extern const char *cp40_get_io_filename(cp40_instance_t *instance);
/*	Return stream filename of INSTANCE or NULL if instance is NULL or
	have no associated file stream. */

extern const char *cp40_get_io_hostname(cp40_instance_t *instance);
/*	Return server hostname of INSTANCE or NULL if instance is NULL or
	have no associated socket stream. */

extern int cp40_get_io_port(cp40_instance_t *instance);
/*	Return server port number of INSTANCE or -1 if instance is NULL or
	have no associated socket stream. */

extern int cp40_get_io_mode(cp40_instance_t *instance);
/*	Return data stream access mode of INSTANCE: CP40_IO_MODE_NONE (no
	stream or INSTANCE is NULL), CP40_IO_MODE_READ, CP40_IO_MODE_WRITE or
	CP40_IO_MODE_UPDATE. */

extern int cp40_get_io_type(cp40_instance_t *instance);
/*	Return data stream type of INSTANCE: CP40_IO_NONE (no stream
	or INSTANCE is NULL), CP40_IO_FILE, CP40_IO_SOCKET. */

extern int cp40_readable(const cp40_instance_t *instance);
/*	Check if INSTANCE has a stream that can be readed in the sense
	that cp40_read_frame can be acted on it. */

extern int cp40_rewindable(const cp40_instance_t *instance);
/*	Check if INSTANCE has a stream that can be rewinded in the sense
	that cp40_rewind_stream can be acted on it. */

extern int cp40_updatable(const cp40_instance_t *instance);
/*	Check if INSTANCE has a stream whose header can be updated in the
	sense that cp40_write_header can be acted on it. */

extern int cp40_writable(const cp40_instance_t *instance);
/*	Check if INSTANCE has a stream that can be written in the sense
	that cp40_write_frame can be acted on it. */

/* Prototype of data stream routines. */
typedef int cp40_read_frame_op_t(cp40_instance_t *instance);
typedef int cp40_write_frame_op_t(cp40_instance_t *instance);
typedef int cp40_read_header_op_t(cp40_instance_t *instance);
typedef int cp40_write_header_op_t(cp40_instance_t *instance);
typedef int cp40_close_stream_op_t(cp40_instance_t *instance);
typedef int cp40_rewind_stream_op_t(cp40_instance_t *instance);
typedef int cp40_set_exposure_op_t(cp40_instance_t *instance, int exposure);

struct _cp40_io_file_t {
  FILE *handle;
  char *name;  /* malloc'ed file name */
};

struct _cp40_io_socket_t {
  /* CP40 socket (only used for reading raw photons and raw frames
     send by, e.g. OS9 photon server. */
  int   handle;  /* socket file descriptor */
  char *host;    /* malloc'ed host name */
  int   port;    /* port number */
  char *buffer;  /* dynamically allocated input data buffer */
  int   size;    /* size of photon buffer */
  int   pending; /* number of bytes stored in input buffer */
};

struct _cp40_stream_t {
  const cp40_stream_ops_t *ops; /* Table of operations on stream (must be the
				   first member). */
  int error; /* Non-zero on read/write error. Each successful attempt to
		read/write stream clear this flag. */
  int end;   /* Non-zero if no new frame could be read (i.g. end of input
		stream or not enough pending input data). Each successful
		attempt to read a new frame clear this flag. */
  union {
    cp40_io_file_t   file;
    cp40_io_socket_t socket;
  } io;
};

/* Structure to hold all stream-dependant operations. */
struct _cp40_stream_ops_t {
  int type; /* Data stream type; one of the following values: */
#define CP40_IO_NONE   0
#define CP40_IO_FILE   1
#define CP40_IO_SOCKET 2

  int mode; /* Data stream access mode; one of the following values: */
#define CP40_IO_MODE_NONE   0 /* No associated stream (must be 0). */
#define CP40_IO_MODE_READ   1 /* Data can be read, parameters can be modified,
				 file header cannot be updated (must be 1). */
#define CP40_IO_MODE_WRITE  2 /* Can write data into a new file, can update
				 file header (must be 2). */
#define CP40_IO_MODE_UPDATE 3 /* Can modify file header and read data. */

  cp40_close_stream_op_t  *close;        /* close data stream */
  cp40_rewind_stream_op_t *rewind;       /* rewind stream to the beginning
					    of data */
  cp40_set_exposure_op_t  *set_exposure; /* set number of frames per
					    super-frame (only effective
					    in read/update modes) */
  cp40_read_frame_op_t    *read_frame;   /* read next raw data frame */
  cp40_write_frame_op_t   *write_frame;  /* write current raw data frame */
  cp40_read_header_op_t   *read_header;  /* (re)read header */
  cp40_write_header_op_t  *write_header; /* (over)write header */
};

/*---------------------------------------------------------------------------*/
/* CP40 DATA INSTANCE */

/* Prototype of photon selection routine. */
typedef int cp40_select_op_t(cp40_frame_t *frame_ptr);

typedef struct _cp40_extra_t cp40_extra_t;

struct _cp40_instance_t {

  /* INFORMATION ABOUT DATA FILE (no incidence on processing) */
  int nphotons[CP40_NCHANNELS];         /* Total photons per channel. */
  int nframes;                          /* Total number of frames. */
  int nsaturations;                     /* Total number of saturations. */

  /* PHOTON BUFFERS FOR THE CURRENT (SUPER-)FRAME */
  cp40_stream_t stream;                 /* Associated data stream and access
					   functions */
  cp40_raw_frame_t *raw_frame;          /* (Super-) frame of raw-photons.
					   Note: super-frames can only be used
					   in read mode of regular data files
					   (i.e., not raw data files). */
  int exposure;                         /* Number of elementary frames per
                                           super-frame, should be >= 1,
                                           irrelevant for write mode. */
  int *count_list;                      /* Counts per elementary raw frames,
                                           has at least EXPOSURE elements
                                           (irrelevant for WRITE mode). */
  cp40_frame_t frame[CP40_NCHANNELS];   /* Cached pre-processed photon-
					   counting frame for every
					   channels. */

  /* TASKS AND REAL-TIME STATISTICS */
  cp40_task_t *task;                    /* Chained list of tasks to perform. */
  int stat_size;                        /* Number of frames to remember. */
  int stat_index;                       /* Current position >=0 and
					   <STAT_SIZE */
  cp40_stat_t *stat;                    /* Array of structures for
					   statistics. */
#define CP40_FOV_REDUCTION 4 /* Shift for real-time FOV task. */
#define CP40_FOV_WIDTH     (2 * (4*CP40_CCD_WIDTH  >> CP40_FOV_REDUCTION))
#define CP40_FOV_HEIGHT    (2 * (4*CP40_CCD_HEIGHT >> CP40_FOV_REDUCTION))
#define CP40_FOV_NBYTES    (CP40_FOV_WIDTH * CP40_FOV_HEIGHT \
                            * sizeof(cp40_task_pixel_t))

  /* SETTINGS FOR DATA PRE-PROCESSING */
  int parity;                           /* Select even/odd numbered frames? */
  cp40_select_op_t *spurious;           /* Deal with spurious photons. */
  int pixel;                            /* Raw-pixel sampling. */
  cp40_select_op_t *correction;         /* Distortion correction method. */
  int notify;                           /* Non-zero if distortion correction
					   parameters (i.e. COEFS, EXTREME,
					   WGHT) must be recomputed due to a
					   change in one of CORRECTION,
					   RAWCOEF, SCALE, ANGLE or PIXEL. */
  double angle;                         /* Rotation angle (in degrees anti-
					   clockwise) after distortion
					   correction. */
  double scale;                         /* Scale after distortion correction
			 		   in pixels per grid step. SCALE=0
					   means no distortion correction. */
  double raw_coefs[CP40_COEF_COUNT];    /* Public (unscaled) set of coefs. */
  double coefs[CP40_COEF_COUNT];        /* Private set of coefficients. */
  cp40_window_t window[CP40_NCHANNELS]; /* Window boundaries as set by user. */
  cp40_window_t extreme[CP40_NCHANNELS];/* Extreme bounds which depend on the
					   distortion correction settings or
					   on the raw-pixel sampling. */
  void *wght[CP40_NCHANNELS];           /* Table for weighted distortion
					   correction. */
  cp40_range_t distance[CP40_NCHANNELS];/* Min/max distance to the edges. */
  cp40_range_t flux[CP40_NCHANNELS];    /* Min/max photon counts. */
  cp40_range_t ratio[CP40_NCHANNELS];   /* Min/max ratio distance/spread. */
  cp40_range_t spread[CP40_NCHANNELS];  /* Min/max image spread. */

  /* EXTRA ATTRIBUTES */
  cp40_hash_table_t  extra_table;       /* Hash table for extra-attributes. */
  cp40_extra_t      *extra_list;        /* List of ordered entries in
					   EXTRA_TABLE. */
  int                extra_count;       /* Number of extra attributes entries
					   in EXTRA_LIST. */
  int                extra_size;        /* Maximum Number of extra attributes
					   entries in EXTRA_LIST. */

  /* HOOK FOR EXTERNAL CLIENT PROGRAMS THAT WOULD IMPORT CP40 FACILITIES */
  void  *client_data;                 /* Anything needed by the client. */
  void (*delete_client_data)(void *); /* If non-NULL, client routine used to
                                         free INSTANCE->client_data when
                                         INSTANCE gets deleted. */
};

/*---------------------------------------------------------------------------*/
/* RAW PHOTON FRAME */

/* Raw  (super-)frame is buffered into cp40_instance_t  structure all channels
   mixed together.   The photons  from a particular  channel are  selected and
   pre-processed   by  cp40_preprocess_photons  and   stored  in   a  struture
   cp40_frame_t. */

struct _cp40_raw_frame_t {
  cp40_raw_photon_t *buffer; /* Dynamically allocated photon buffer. */
  int date;                  /* Frame number. */
  int count;                 /* Actual number of photons in buffer. */
  int size;                  /* Maximum number of photons in buffer. */
#define CP40_RAW_FRAME_STATIC_BUFFER_SIZE 1020
  cp40_raw_photon_t static_buffer[CP40_RAW_FRAME_STATIC_BUFFER_SIZE];
};

extern cp40_raw_frame_t *cp40_new_raw_frame(void);
/*	Allocate and instanciate a new FRAME.  Return NULL in case of error. */

extern void cp40_delete_raw_frame(cp40_raw_frame_t *frame);
/*	Free FRAME and its resources. */

void cp40_reset_raw_frame(cp40_raw_frame_t *frame);
/*	Reset raw photon buffer of FRAME to be empty. */

extern int cp40_resize_raw_frame(cp40_raw_frame_t *frame, int new_size);
/*	Set size of raw photon buffer of FRAME to be NEW_SIZE.  The operation
	should be protected against interrupts unless compiler optimizations
	change the order of operations.  A standard result is returned: CP40_OK
	or CP40_ERROR.  Note: if NEW_SIZE is smaller or equal the size of the
	static buffer embedded into FRAME, then this static buffer gets used
        and FRAME->size becomes the size of the static buffer. */

/* PROTECTION AGAINST INTERRUPTS
     A routine protected against interrupts can be safely interrupted (e.g. by
     a signal) because it takes care to keep everything in a consistent state.
     The only damage an interrupt can cause is the loss of some dyamically
     allocated buffer. */
/*---------------------------------------------------------------------------*/

/*_____________________________________________________________________________
 *
 * Basic operations
 * ================
 *
 *	You must first create a new CP40 data instance (i.e. allocate and
 *	instanciate with defaults a new cp40_instance_t structure) with
 *	cp40_new_instance(); e.g.:
 *
 *	    cp40_instance_t *instance = cp40_new_instance();
 *	    if (instance == NULL) {
 *		fprintf(stderr, "%s\n", cp40_error_info(1));
 *		exit(3);
 *	    }
 *
 *	At this stage, you can only get or change attributes of the new
 *	CP40 data instance (see section "Data attributes").  In order to
 *	perform input/output operation (that is reading or writing photons),
 *	you must associate the CP40 data instance with a stream with
 *	functions: cp40_assoc_file() or cp40_assoc_socket().  The input/output
 *	operations are described in section "Data input/output and
 *	pre-processing".
 *
 *	When a CP40 data instance has an associated stream, you can use
 *	cp40_read_header() to read or read again CP40 data file header, e.g.
 *	to revert attribute changes.  cp40_read_header() is automatically
 *	called when a data stream is open in read or update mode to initialize
 *	attribute values.  Besides, cp40_write_header() (re)write CP40 data
 *	stream header, e.g. to update attribute changes.  cp40_write_header()
 *	may be automatically called when closing a data stream which was open
 *	in write or update mode.
 *
 *	When the CP40 data instance is no more needed, you just call
 *	cp40_delete_instance(instance, 1/0) to release all related resources
 *	(if second argument is non-zero, cp40_delete_instance takes care of
 *	updating header if this is relevant).
 */
extern cp40_instance_t *cp40_new_instance(void);
extern int cp40_delete_instance(cp40_instance_t *instance, int update);
extern int cp40_get_file_version(const char *path);

/*---------------------------------------------------------------------------*/
/* HIGH-LEVEL INTERFACE FOR INSTANCE ATTRIBUTES */

/* CP40 library provides you with two levels of interface routines to manage
   CP40 attributes.

   The  low-level  interface  routines  query  or  set  *STANDARD*  attributes
   (i.e. those in cp40_instance_t  structure) by value.  Since some attributes
   have side effects (e.g.  modifying the distortion correction method trigger
   the  updating of  distortion  correction parameters)  and  some others  are
   read-only,  the user  should use  these routines  preferably  than directly
   setting/getting  values in  the cp40_instance_t  structure.   The interface
   routines handle these underlying side effects.

   By  contrast, the  high-level  interface routines  handle *ALL*  attributes
   (i.e. the standard ones plus the extra ones) by human-readable keywords and
   string values.  In effect, they reference all attributes by their names and
   they convert their values to/from human readable strings.  Extra-attributes
   all  have string  values  (i.e. no  binary  counterpart in  cp40_instance_t
   structure)  and  are  stored the  private  hash  table  of each  CP40  data
   instance.  Only  two routines are  sufficient to implement this  high level
   interface (plus a further one to get the name and value of extra attributes
   in the same order as they were created). */

extern int cp40_get(cp40_instance_t *instance, const char *name,
		    cp40_dstr_t *value);
/* Get  human-readable string value of  attribute NAME for  CP40 data INSTANCE
   and store  it into  dynamic string VALUE.   VALUE must have  been correctly
   initialized by cp40_dstr_init.  Return standard result: CP40_OK or CP40_ERROR (error
   message can be retrieved by cp40_error_info). */

extern int cp40_set(cp40_instance_t *instance, const char *name,
		    const char *value);
/* Use human-readable string VALUE to set the value of attribute NAME for CP40
   data INSTANCE.  Return standard result: CP40_OK or CP40_ERROR (error message
   can be retrieved by cp40_error_info). */

extern int cp40_get_next_extra(cp40_instance_t *instance, int *index_ptr,
			       const char **name_ptr, const char **value_ptr);
/*  Get  name and  value  of  next extra  attribute  for  CP40 data  INSTANCE.
   INDEX_PTR is the  address where is stored index  of current extra attribute
   (contents  should be initialized  to a  negative value  to get  first extra
   attribute) also used to store  index of next extra attribute.  NAME_PTR and
   VALUE_PTR  are addresses  where to  store  next attribute  name and  value.
   Return value  is 1 upon success,  0 if no more  extra attributes (*NAME_PTR
   and *VALUE_PTR set to NULL) or -1 in case of error. */

/*---------------------------------------------------------------------------*/
struct _cp40_extra_t {
  /* Structure to store extra (i.e.  non standard) attribute value (to be used
     in  conjonction with a  hash table  so as  to combine  the features  of a
     reference-by-keyword table and reference-by-order list).  */
  cp40_hash_entry_t *entry;     /* Corresponding entry in hash table. */
  const char *string;	 /* Address of dynamic memory to store the extra
			    attribute value. */
};

struct _cp40_attr_entry_t {
  /* An attribute table is a  cp40_hash_table indexed by label strings.  The value of
     each entry  in this  ``label table'' is  a malloc'ed cp40_hash_table  indexed by
     attribute names.  The  value of each entry in these  ``name tables'' is a
     malloc'ed cp40_attr_entry structure and the corresponding attribute value
     is a malloc'ed string. */
  cp40_hash_entry_t *label;	/* Attribute label. */
  cp40_hash_entry_t *name;	/* Attribute name. */
  const char *value;	/* Attribute value (malloc'ed string). */
};

/* The following macros take a pointer to a valid cp40_attr_entry. */
#define cp40_get_attr_value(ptr)  ((ptr)->value)
#define cp40_get_attr_name(ptr)   ((ptr)->name->key.string)
#define cp40_get_attr_label(ptr)  ((ptr)->label->key.string)
#define cp40_init_attr_table(ptr) cp40_hash_init_table(ptr, CP40_HASH_STRING_KEYS)

extern cp40_attr_entry_t *cp40_find_attr_entry(cp40_hash_table_t *label_table,
					       const char *label,
					       const char *name);
extern cp40_attr_entry_t *cp40_create_attr_entry(cp40_hash_table_t *label_table,
						 const char *label,
						 const char *name,
						 int *new_ptr);
extern void cp40_delete_attr_entry(cp40_attr_entry_t *attr_entry);
extern void cp40_delete_attr_table(cp40_hash_table_t *label_table);
extern int  cp40_set_attr_value(cp40_attr_entry_t *attr_entry,
				const char *value);
extern int  cp40_create_attr_entry_with_value(cp40_hash_table_t *lab_table,
					      const char *label,
					      const char *name,
					      const char *value);

/* Macro version of cp40_create_attr_entry_with_value. */
#define cp40_create_attr_entry_with_value(table, label, name, value) \
    cp40_set_attr_value(cp40_create_attr_entry(table, label, name, NULL), value)

/* The following is to save/restore attribute settings. */
#define CP40_ATTR_STANDARD        (1<<0) /* Save/restore only standard
					    attributes (Note: read only
					    attributes are ignored or get
					    commented). */
#define CP40_ATTR_EXTRA           (1<<1) /* Save/restore only extra (non-
					    standard) attributes. */
#define CP40_ATTR_PREPROCESSING   (1<<2) /* Save/restore only attributes
					    involved in pre-processing. */
#define CP40_ATTR_ALL             (CP40_ATTR_STANDARD|CP40_ATTR_EXTRA)

extern int cp40_save_attr(cp40_instance_t *instance, const char *filename,
			  int flags);
extern int cp40_restore_attr(cp40_instance_t *instance, const char *filename,
			     int flags);
extern int cp40_copy_attr(cp40_instance_t *dst_ptr,
	      /* const */ cp40_instance_t *src_ptr, int flags);

/*---------------------------------------------------------------------------*/

#define CP40_UPDATE_CACHE_MAYBE(ip) if ((ip)->notify) cp40_update_cache(ip)

extern void cp40_update_cache(cp40_instance_t *instance);
/* (Re)compute distortion correction settings and frame boundaries.  Must be
   called before any pre-processing operation as soon as INSTANCE->notify
   is non-zero.  The routine computes the values of:
     INSTANCE->coefs
     INSTANCE->extreme
     INSTANCE->frame[0..3].xmin (as well as xmax, ymin and ymax)
   according to the values of:
     INSTANCE->raw_coefs
     INSTANCE->pixel
     INSTANCE->correction
     INSTANCE->angle
     INSTANCE->scale
     INSTANCE->window[0..3]
   and set INSTANCE->notify to zero.   If you want to force computation
   of settings do:
     INSTANCE->notify= 3;
     cp40_update_cache(INSTANCE);
*/

extern int cp40_get_nframes(cp40_instance_t *instance);
/* Get total number of frames in a CP40 data file. Return INSTANCE->nframes
   or -1 for invalid argument. */

extern int cp40_get_nsaturations(cp40_instance_t *instance);
/* Get total number of saturations in a CP40 data file. Return
   INSTANCE->nsaturations or -1 for invalid argument. */

extern int cp40_get_nphotons(cp40_instance_t *instance, int channel);
/* Get total number of photons in given channel of a CP40 data file.
   Return INSTANCE->nphotons[CHANNEL] or -1 for invalid argument. */

extern int cp40_get_frame_count(cp40_instance_t *instance);
/* Get total number of photons in current (super-)frame buffer of a CP40
   data file.  Return INSTANCE->raw_frame->count or -1 for invalid argument. */

extern int cp40_get_frame_date(cp40_instance_t *instance);
/* Get total date (number) of current (super-)frame buffer of a CP40
   data file.  Return INSTANCE->raw_frame->date or -1 for invalid argument. */

extern int cp40_get_exposure(cp40_instance_t *instance);
/* Get current number of short-exposure frames per super-frame.  Return
   INSTANCE->exposure or -1 for invalid argument. */

/*---------------------------------------------------------------------------*/

extern int cp40_get_limits(cp40_instance_t *instance, int channel,
			   int *xmin, int *xmax, int *ymin, int *ymax);
/*  Get  coordinate boundaries  after  pre-processing.  Distortion  correction
   settings get updated if needed. */

extern int cp40_get_limits2(cp40_instance_t *instance, int chn1, int chn2,
			    int *xmin, int *xmax, int *ymin, int *ymax);
/* Get coordinate boundaries after pre-processing for an auto/intercorrelation
   between CHN1 and CHN2.  Positions in  the result are taken as: POSITION_2 -
   POSITION_1 (i.e. CHN1 is the origin).  Return standard result. */


extern int cp40_set_window(cp40_instance_t *instance, int channel,
			   cp40_window_t *window_ptr);
extern int cp40_get_window(cp40_instance_t *instance, int channel,
			   cp40_window_t *window_ptr);
extern int cp40_get_extreme(cp40_instance_t *instance, int channel,
			    cp40_window_t *window_ptr);
extern int cp40_set_distance(cp40_instance_t *instance, int channel,
			     cp40_range_t *range_ptr);
extern int cp40_get_distance(cp40_instance_t *instance, int channel,
			     cp40_range_t *range_ptr);
extern int cp40_set_flux(cp40_instance_t *instance, int channel,
			 cp40_range_t *range_ptr);
extern int cp40_get_flux(cp40_instance_t *instance, int channel,
			 cp40_range_t *range_ptr);
extern int cp40_set_ratio(cp40_instance_t *instance, int channel,
			  cp40_range_t *range_ptr);
extern int cp40_get_ratio(cp40_instance_t *instance, int channel,
			  cp40_range_t *range_ptr);
extern int cp40_set_spread(cp40_instance_t *instance, int channel,
			   cp40_range_t *range_ptr);
extern int cp40_get_spread(cp40_instance_t *instance, int channel,
			   cp40_range_t *range_ptr);


/*_____________________________________________________________________________
 *
 * Data input/output and pre-processing
 * ====================================
 *
 *	The raw-photons from the current (super-)frame are directly stored
 *	into the corresponding cp40_instance_t structure.  In this internal
 *	buffer raw-photons from all channels are mixed together.  The macro
 *	cp40_read_frame() read next raw photon frame from data stream and
 *	store it into the internal super-frame buffer (cp40_instance_t must be
 *	`readable'; use CP40_Readable() function to check this).  The macro
 *	cp40_write_frame() triggers the writting of the internal raw-photon
 *	buffer into the data stream as a single raw frame (i.e. no
 *	super-frame support is provided in write/update mode).  In order to
 *	ensure that the current raw-photon buffer is large enough to store
 *	your raw-photons, use cp40_set_size with the number of events you
 *	intend to store in instance->buffer.  */
#define cp40_write_frame(ptr)    ((ptr)->stream.ops->write_frame(ptr))
extern int cp40_read_frame(cp40_instance_t *instance);
/*	Read next raw photon (super-)frame from CP40 data stream of INSTANCE.
	Possible return values are:
	   1 if a new frame was effectively read
	   0 if no new frame could be read (either because normal
	     end-of-stream or not enough pending input data).
	  -1 on error.
	You can also check the flags INSTANCE->io_error and INSTANCE->io_end
	for a diagnostic. */


/*	Processed photons (i.e. no more "raw" ones) are stored in the dynamic
 *	photon buffer of a cp40_frame_t structure.  A cp40_frame_t structure is
 *	created by a call to cp40_new_frame().  Before any use of a
 *	cp40_frame_t structure, you must configure it by a call to
 *	cp40_configure_frame() THIS IS MANDATORY to insure proper settings.
 *	You should not modify preprocessing options after
 *	cp40_configure_frame() since they affects the limits stored in the
 *	cp40_frame_t structure nor should you modify contents of frame_ptr.
 *	In order to select photons from the channel of interest and to
 *	apply pre-processing (as specified by attributes) you just call the
 *	function cp40_preprocess_photons() (which also instanciates statistics in
 *	the argument cp40_frame_t structure).  When you no more need the
 *	cp40_frame_t structure, you must free its internal buffer(s) with
 *	cp40_delete_frame().  You don't need to free your cp40_frame_t
 *	structure between calls to cp40_preprocess_photons() (even if you use it
 *	to get photons from a different channel) nor you need to set the
 *	size of its internal buffer with the function cp40_resize_frame()
 *	this is automatically done by cp40_preprocess_photons().
 *
 *	It is guaranted that photons stored in a cp40_frame_t structure
 *	by cp40_preprocess_photons() have been properly selected; in particular
 *	their positions are within the bounds returned by cp40_get_limits()
 *	(inclusive).  */

extern int cp40_import(cp40_instance_t *instance,
		       const cp40_raw_photon_t *buffer,
		       long number, int flags);
/*	Import `number` raw photons from `buffer` into writable CP40 stream
	`instance`.  Photon with highest bit cleared (set to 0) are used
	between frames: every positive photon marks the end/beginning of a
	frame if bit CP40_IMPORT_LAST is non-zero/zero in `flags`.  If bit
	CP40_IMPORT_WFS of `flags` is set, the imported frames get numbered
	according to the wavefront sensor number.   Frames can span
	across input buffer boundaries.  The routine returns ETL_OK or
	ETL_ERROR. */
#define CP40_IMPORT_LAST 0x1
#define CP40_IMPORT_WFS  0x2


/*_____________________________________________________________________________
 *
 * Integration
 * ===========
 *
 *	cp40_stat	Integrate frame statistics.
 *	cp40_it		Integrate (long exposure) image.
 *	cp40_ac		Integrate autocorrelation.
 *	cp40_ic		Integrate intercorrelation.
 *
 *	For all these functions: COUNT is the number of frames to read (if
 *	COUNT=0, photons from current super-frame get used);  NSELPTR is the
 *	address to store number of selected frames (even in case of error,
 *	you can trust this number);  WIDTH and HEIGHT are the dimensions of
 *	output array ARRAYPTR (for the cp40_stat function ARRAYPTR must have
 *	at least 7 x MAX(COUNT,1) elements); other arguments have obvious
 *	meaning.  The returned value is: 0 on end-of-data, -1 on error, and
 *	1 otherwise.  Pre-processing and frame selection are applied as
 *	indicated by the attributes of DATAPTR (with one exception: if
 *	COUNT=0 and if you have changed these attributes after the last
 *	previous frame selection).  These functions make use of 1 or 2
 *	private cp40_frame_t structures to gather photons (you don't have to
 *	worry about that).
 */
#define CP40_ALL_FRAMES  (2147483647L)	/* A hopefully large value. */

extern int cp40_stat(cp40_instance_t *instance, double *array_ptr,
		     long count, int chn, long *n_sel_ptr);

extern int cp40_it(cp40_instance_t *instance, long *array_ptr, long width,
		   long height, long count, int chn, long *n_sel_ptr);

extern int cp40_ac(cp40_instance_t *instance, long *array_ptr, long width,
		   long height, long count, int chn, long *n_sel_ptr,
		   int flags);

extern int cp40_ic(cp40_instance_t *instance, long *array_ptr, long width,
		   long height, long count, int chn1, int chn2,
		   long *n_sel_ptr);

/*---------------------------------------------------------------------------*/
/* TASKS AND REAL-TIME STATISTICS */
typedef double cp40_task_pixel_t; /* Data type for integration of CP40
				     estimators. */
struct _cp40_task_t {
  cp40_instance_t *owner;      /* This is needed for backward reference. */
  cp40_task_t *next;           /* Point to next task or NULL. */
#define CP40_TOKEN_SIZE 8      /* Maximum size (length+1) for task token
                                  string (8 should be enough). */
  char token[CP40_TOKEN_SIZE]; /* Task identifier. */
  cp40_task_op_t *op;          /* The operation to execute. */
  int run;                     /* Run task? */
  int chn1, chn2;              /* Photon-counting channels (<0 if
                                  irrelevant). */
  int width, height;           /* Dimensions of result array. */
  int xmin, xmax, ymin, ymax;  /* Boundaries of result array. */
  cp40_task_pixel_t **pixel;   /* Array of row pointers for fast indexing
                                  of 2-D array RESULT, e.g., just do:
                                    int **ptr= task->pixel;
                                    ptr[y][x]= ...; */
  cp40_task_pixel_t data[1];   /* Image to store the result. _MUST_ be the
				  last member: actual size will be sufficient
				  to store the result. */
};

struct _cp40_stat_t {
  cp40_raw_frame_t *raw_frame;
  double flux[CP40_NCHANNELS];  /* Flux for each channel after pre-processing. */
  double x[CP40_NCHANNELS];     /* Mean abscissae for ... */
  double y[CP40_NCHANNELS];     /* Mean ordinate for ... */
  double spread[CP40_NCHANNELS];/* Mean image spread for ... */
};

extern cp40_task_op_t *cp40_parse_task_token(const char *token,
					     int *chn1, int *chn2);
extern char           *cp40_print_task_token(char *token, const char *task,
					     int chn1, int chn2);
/* Parse/format task token from/into TOKEN (at least 5 bytes).
   cp40_parse_task_token() return address of task operation routine and store
   channel number(s) (or -1) in CHN1 and CHN2 (if not NULL).
   cp40_print_task_token() return TOKEN; both routines return NULL in case of
   error.*/

extern cp40_task_t *cp40_schedule_task(cp40_instance_t *instance,
				       const char *token);
/* Schedule a photon-counting task to run on data from INSTANCE.  The task
   type and related channels is given by TOKEN (see cp40_parse_task_token).
   If the task is already scheduled, its structure address is returned;
   otherwise resources for a new task are allocated and the address of the new
   stack structure is returned.  In case of error NULL is returned.  Note:
   INSTANCE keep track of its tasks so there is no need to remember the task
   structure address. */

extern cp40_task_t *cp40_get_task(cp40_instance_t *instance,
				  const char *token);
/* Return structure address (NULL in case of error) of task scheduled in
   INSTANCE and identified by TOKEN (see cp40_parse_task_token). */

extern void cp40_stop_task(cp40_task_t *task);
/* Stop a running task (its result is still visible).  Nothing appends if TASK
   is NULL or is already stopped. */

extern void cp40_start_task(cp40_task_t *task);
/* Start a stopped photon-counting task.  Nothing appends if TASK is NULL or
   already running. */

extern void cp40_cancel_task(cp40_task_t *task);
/* Cancel a photon-counting task. Free all related resources and the task is
   no longer existing.  Nothing appends if TASK is NULL. */

extern void cp40_clear_task(cp40_task_t *task);
/* Set all TASK pixels to zero. */

extern void cp40_start_all_tasks(cp40_instance_t *instance);
extern void cp40_stop_all_tasks(cp40_instance_t *instance);
extern void cp40_cancel_all_tasks(cp40_instance_t *instance);
extern void cp40_clear_all_tasks(cp40_instance_t *instance);
/*  Apply global task operation for all scheduled tasks of INSTANCE. */

extern int cp40_set_stat_duration(cp40_instance_t *instance, int size);
/* Set statistics duration to be SIZE frames for INSTANCE.  If SIZE<=0, free
   all resources related to integration of statistics.  This is intended for
   real-time operation. */

extern int cp40_run_tasks(cp40_instance_t *instance, int count);
/* If COUNT>=0, read and process as explained below at most that number of
   frames (not accounting for discarded frames); otherwise, i.e.  COUNT<0,
   read and process all remaining frames.  The number of processed frames (not
   accounting for discarded frames) is returned.  The caller can check
   INSTANCE->io_error for possible errors.  Real-time processes can safely
   ignore such errors.

   For each frame, the processing consists in:
     1. reading next raw (super-)frame;
     2. apply preprocessing;
     3. update statistics for real-time;
     4. update real-time field-of-view image;
     5. stole raw (super-)frame (this is needed for circular averaging
        involved in real-time operation). */

#define CP40_USE_NAME_AS_PREFIX (1<<0)
#define CP40_MAY_OVERWRITE      (1<<1)
extern int cp40_save_task(cp40_task_t *task, const char *name, int flags);
/* Set TASK image as FITS file NAME.  If bit CP40_USE_NAME_AS_PREFIX of FLAGS
   is set, the output file name will be: "NAME-TOKEN.fits"; where TOKEN is the
   task name.  Return standard result: CP40_OK or CP40_ERROR. */

struct _cp40_avg_stat_t {
  int nframes; /* sample size */
  double mean_flux, mean_x, mean_y, mean_spread;
};

extern int cp40_average_stat(cp40_instance_t *instance,
			     cp40_avg_stat_t avg[CP40_NCHANNELS]);
/* Fill AVG with average statistics from real-time information collected in
   INSTANCE (see cp40_run_tasks).  Return sample size (indication of the
   number of frames accounted for). */

/*---------------------------------------------------------------------------*/
/* MISCELLANEOUS */

extern int cp40_check_channel(int channel);
/* Check CP40 channel number. Return standard result: CP40_OK or CP40_ERROR.  In
   case of error, left an error message (see cp40_error_info). */

/*---------------------------------------------------------------------------*/
/* PRIVATE RESOURCES/ROUTINES */

extern void _cp40_invalidate_cache(cp40_instance_t *instance);
/* Invalidate computed distortion correction settings to force them to be
   recomputed the next time they are needed.  All related resources
   (e.g. tables for weighted correction) get freed.  This affect: effective
   distortion correction coefficients, weighted correction tables, limits.
   WARNING: before calling this routine, make sure that there are no pending
   integration, in particular no scheduled tasks (i.e. INSTANCE->task must
   be NULL). */

extern void _cp40_init_frame(cp40_frame_t *frame,
			     cp40_instance_t *owner, int channel);
/* Instanciate photon-counting frame.  Must be called before first use.
   Must not be called again before internal buffers get freed by
   _cp40_resize_frame(FRAME, 0). */

extern int _cp40_resize_frame(cp40_frame_t *frame, int size);
/* (Re)alloc photon buffer of FRAME to be at least SIZE photons (SIZE be
   greater or equal actual FRAME->count).  Return standard result: CP40_OK or
   CP40_ERROR -- in case of error, a message is left that can be retrieved by a
   call to cp40_error_info().  If SIZE is 0, intenal buffers get freed. */

/*---------------------------------------------------------------------------*/
_CP40_END_DECLS
#endif /* _CP40_H */
