/*
 * cp40_hash.c --
 *
 *	Implementation of in-memory hash tables mostly based on
 *	implementation made in Tcl with some improvements (take care of
 *	memory allocation failures, const char keys, use of INLINE
 *	declaration, use of `delete' method for hash value...).
 *
 * Copyright (c) 1991-1993 The Regents of the University of California.
 * Copyright (c) 1994 Sun Microsystems, Inc.
 *	See the file "LICENSE.TCL" for information on usage and
 *	redistribution of Tcl sources and for a DISCLAIMER OF ALL
 *	WARRANTIES.
 *
 * Copyright (c) 1995-2006 Eric Thibaut.
 *	See the file "COPYING" for information on usage and redistribution
 *	of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 *-----------------------------------------------------------------------------
 *
 * History:
 *	$Id$
 *	$Log$
 */

#include <stdlib.h>
#include <string.h>

#include "cp40.h"

/*---------------------------------------------------------------------------*/

/* When there are this many entries per bucket, on average, rebuild the hash
   table to make it larger. */
#define REBUILD_MULTIPLIER  3


/* The following macro takes a preliminary integer hash value and produces an
   index into a hash tables bucket list.  The idea is to make it so that
   preliminary values that are arbitrarily similar will end up in different
   buckets.  The hash function was taken from a random-number generator. */
#define RANDOM_INDEX(table, i) \
    (((((long) (i))*1103515245) >> (table)->down_shift) & (table)->mask)


/*
 * Procedure prototypes for static procedures in this file:
 */

static unsigned int hash_string(const unsigned char *string);
static void rebuild_table(cp40_hash_table_t *table);
static cp40_hash_entry_t *array_find(cp40_hash_table_t *table,
				     const char *key);
static cp40_hash_entry_t *array_create(cp40_hash_table_t *table,
				       const char *key, int *new);
static cp40_hash_entry_t *bogus_find(cp40_hash_table_t *table,
				     const char *key);
static cp40_hash_entry_t *bogus_create(cp40_hash_table_t *table,
				       const char *key, int *new);
static cp40_hash_entry_t *string_find(cp40_hash_table_t *table,
				      const char *key);
static cp40_hash_entry_t *string_create(cp40_hash_table_t *table,
					const char *key, int *new);
static cp40_hash_entry_t *one_word_find(cp40_hash_table_t *table,
					const char *key);
static cp40_hash_entry_t *one_word_create(cp40_hash_table_t *table,
					  const char *key, int *new);

static const cp40_hash_ops_t array_ops= {array_find, array_create};
static const cp40_hash_ops_t bogus_ops= {bogus_find, bogus_create};
static const cp40_hash_ops_t string_ops= {string_find, string_create};
static const cp40_hash_ops_t one_word_ops= {one_word_find, one_word_create};

/*_____________________________________________________________________________
 *
 * cp40_hash_init_table --
 *
 *	Given storage for a hash table, set up the fields to prepare
 *	the hash table for use.
 *
 * Arguments:
 *	TABLE - Pointer to table record, which is supplied by the caller.
 *	KEY_TYPE - Type of keys to use in table: CP40_HASH_STRING_KEYS,
 *		CP40_HASH_ONE_WORD_KEYS, or an integer >= 2.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	TablePtr is now ready to be passed to cp40_hash_find_entry and
 *	cp40_hash_create_entry.
 *_____________________________________________________________________________
 */

void cp40_hash_init_table(cp40_hash_table_t *table, int key_type)
{
  table->ops= &bogus_ops; /* will panic if KEY_TYPE is invalid */
  table->buckets= table->static_buckets;
  table->static_buckets[0]= 0;
  table->static_buckets[1]= 0;
  table->static_buckets[2]= 0;
  table->static_buckets[3]= 0;
  table->num_buckets= CP40_HASH_SMALL_TABLE;
  table->num_entries= 0;
  table->rebuild_size= CP40_HASH_SMALL_TABLE * REBUILD_MULTIPLIER;
  table->down_shift= 28;
  table->mask= 3;
  table->key_type= key_type;
  if (key_type == CP40_HASH_STRING_KEYS) {
    table->ops= &string_ops;
  } else if (key_type == CP40_HASH_ONE_WORD_KEYS) {
    table->ops= &one_word_ops;
  } else if (key_type >= 2) {
    table->ops= &array_ops;
  }
}

/*_____________________________________________________________________________
 *
 * cp40_hash_delete_entry --
 *
 *	Remove a single entry from a hash table.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	The entry given by entry_ptr is deleted from its table and
 *	should never again be used by the caller.  Call ENTRY->delete
 *	to free ENTRY->client_data if such a callback was registered,
 *	otherwise it is up to the caller to free the ENTRY->client_data
 *	member of the entry, if that is relevant.
 *_____________________________________________________________________________
 */

void cp40_hash_delete_entry(cp40_hash_entry_t *entry)
{
  cp40_hash_entry_t *previous;

  if (entry == NULL) {
    return;
  }

  /* Remove entry from hash table. */
  if (*entry->bucket_ptr == entry) {
    *entry->bucket_ptr= entry->next_ptr;
  } else {
    for (previous = *entry->bucket_ptr ; ; previous = previous->next_ptr) {
      if (previous == NULL) {
	cp40_fatal_error("invalid bucket chain in routine cp40_hash_delete_entry (" __FILE__ ":" CP40_STRINGIFY(__LINE__));
      }
      if (previous->next_ptr == entry) {
	previous->next_ptr= entry->next_ptr;
	break;
      }
    }
  }
  --entry->table->num_entries;

  /* delete related ressources. */
  if (entry->delete != NULL) entry->delete(entry->client_data);
  free(entry);
}

/*_____________________________________________________________________________
 *
 * cp40_hash_set_value --
 *
 *	Set value of a hash entry.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	The contents of entry_ptr is deleted and its new value is set.
 *_____________________________________________________________________________
 */

void cp40_hash_set_value(cp40_hash_entry_t *entry, void *value,
			 void (*delete)(void *))
{
  void (*old_delete)(void *)= entry->delete;
  void *old_value= entry->client_data;
  entry->client_data= value;
  entry->delete= delete;
  if (old_delete != NULL) old_delete(old_value);
}

/*_____________________________________________________________________________
 *
 * cp40_hash_delete_table --
 *
 *	delete up everything associated with a hash table except for
 *	the record for the table itself.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	The hash table is no longer useable.
 *_____________________________________________________________________________
 */

void cp40_hash_delete_table(cp40_hash_table_t *table)
{
  cp40_hash_entry_t *entry, *next_ptr;
  int i;

  /* Arrange for panics if the table is used again without
     re-initialization. */
  table->ops= &bogus_ops;

  /* Delete all the entries in the table. */
  for (i=0 ; i < table->num_buckets ; i++) {
    entry= table->buckets[i];
    while (entry != NULL) {
      next_ptr= entry->next_ptr;
      if (entry->delete != NULL) entry->delete(entry->client_data);
      free(entry);
      entry= next_ptr;
    }
  }

  /* Delete the bucket array, if it was dynamically allocated. */
  if (table->buckets != table->static_buckets) {
    free(table->buckets);
  }
}

/*_____________________________________________________________________________
 *
 * cp40_hash_first_entry --
 *
 *	Locate the first entry in a hash table and set up a record
 *	that can be used to step through all the remaining entries
 *	of the table.
 *
 * Results:
 *	The return value is a pointer to the first entry in table,
 *	or NULL if table has no entries in it.  The memory at
 *	*search_ptr is initialized so that subsequent calls to
 *	cp40_hash_next_entry will return all of the entries in the table,
 *	one at a time.
 *
 * Side effects:
 *	None.
 *_____________________________________________________________________________
 */

cp40_hash_entry_t *cp40_hash_first_entry(cp40_hash_table_t *table,
					 cp40_hash_search_t *search)
{
  search->table= table;
  search->next_index= 0;
  search->next_entry= NULL;
  return cp40_hash_next_entry(search);
}

/*_____________________________________________________________________________
 *
 * cp40_hash_next_entry --
 *
 *	Once a hash table enumeration has been initiated by calling
 *	cp40_hash_first_entry, this procedure may be called to return
 *	successive elements of the table.
 *
 * Arguments:
 *	SEARCH - Place to store information about progress through the
 *	         table.  Must have been initialized by calling
 *	         cp40_hash_first_entry.
 *
 * Results:
 *	The return value is the next entry in the hash table being
 *	enumerated, or NULL if the end of the table is reached.
 *
 * Side effects:
 *	None.
 *_____________________________________________________________________________
 */

cp40_hash_entry_t *cp40_hash_next_entry(cp40_hash_search_t *search)
{
  cp40_hash_entry_t *entry;

  while (search->next_entry == NULL) {
    if (search->next_index >= search->table->num_buckets) {
      return NULL;
    }
    search->next_entry=
      search->table->buckets[search->next_index];
    search->next_index++;
  }
  entry= search->next_entry;
  search->next_entry= entry->next_ptr;
  return entry;
}

/*_____________________________________________________________________________
 *
 * cp40_hash_stats --
 *
 *	Return statistics describing the layout of the hash table
 *	in its hash buckets.
 *
 * Results:
 *	The return value is a malloc-ed string containing information
 *	about table.  It is the caller's responsibility to free
 *	this string.
 *
 * Side effects:
 *	None.
 *_____________________________________________________________________________
 */

char *cp40_hash_stats(cp40_hash_table_t *table)
{
#define NUM_COUNTERS 10
  int count[NUM_COUNTERS], overflow, i, j;
  double average, tmp;
  cp40_hash_entry_t *hptr;
  char *result, *p;

  /*
   * Compute a histogram of bucket usage.
   */

  for (i= 0; i < NUM_COUNTERS; i++) {
    count[i]= 0;
  }
  overflow= 0;
  average= 0.0;
  for (i= 0; i < table->num_buckets; i++) {
    j= 0;
    for (hptr= table->buckets[i]; hptr != NULL; hptr= hptr->next_ptr) {
      j++;
    }
    if (j < NUM_COUNTERS) {
      count[j]++;
    } else {
      overflow++;
    }
    tmp= j;
    average += (tmp+1.0)*(tmp/table->num_entries)/2.0;
  }

  /*
   * Print out the histogram and a few other pieces of information.
   */

  result= (char *) malloc((unsigned) ((NUM_COUNTERS*60) + 300));
  if (result == NULL) {
    cp40_error("not enough memory for hash table statistics");
    return NULL;
  }
  sprintf(result, "%d entries in table, %d buckets\n",
	  table->num_entries, table->num_buckets);
  p= result + strlen(result);
  for (i= 0; i < NUM_COUNTERS; i++) {
    sprintf(p, "number of buckets with %d entries: %d\n",
	    i, count[i]);
    p += strlen(p);
  }
  sprintf(p, "number of buckets with %d or more entries: %d\n",
	  NUM_COUNTERS, overflow);
  p += strlen(p);
  sprintf(p, "average search distance for entry: %.1f", average);
  return result;
}

/*_____________________________________________________________________________
 *
 * hash_string --
 *
 *	Compute a one-word summary of a text string, which can be
 *	used to generate a hash index.
 *
 * Arguments:
 *	STRING - String from which to compute hash value.
 *
 * Results:
 *	The return value is a one-word summary of the information in
 *	string.
 *
 * Side effects:
 *	None.
 *_____________________________________________________________________________
 */

static __inline__ unsigned int hash_string(const unsigned char *string)
{
  unsigned int c, result = 0;

  /*
   * I tried a zillion different hash functions and asked many other
   * people for advice.  Many people had their own favorite functions,
   * all different, but no-one had much idea why they were good ones.
   * I chose the one below (multiply by 9 and add new character)
   * because of the following reasons:
   *
   * 1. Multiplying by 10 is perfect for keys that are decimal strings,
   *    and multiplying by 9 is just about as good.
   * 2. Times-9 is (shift-left-3) plus (old).  This means that each
   *    character's bits hang around in the low-order bits of the
   *    hash value for ever, plus they spread fairly rapidly up to
   *    the high-order bits to fill out the hash value.  This seems
   *    works well both for decimal and non-decimal strings.
   */

  while ((c = *string++) != 0) {
    result += (result<<3) + c;
  }
  return result;
}

/*_____________________________________________________________________________
 *
 * string_find --
 *
 *	Given a hash table with string keys, and a string key, find
 *	the entry with a matching key.
 *
 * Arguments:
 *	TABLE - Table in which to lookup entry.
 *	KEY   - Key to use to find matching entry.
 *
 * Results:
 *	The return value is a token for the matching entry in the
 *	hash table, or NULL if there was no matching entry.
 *
 * Side effects:
 *	None.
 *_____________________________________________________________________________
 */

static cp40_hash_entry_t *string_find(cp40_hash_table_t *table,
				      const char *key)
{
  cp40_hash_entry_t *hptr;
  const char *p1, *p2;
  size_t index;

  if (key == NULL) return NULL;
  index = hash_string((const unsigned char *)key) & table->mask;

  /*
   * Search all of the entries in the appropriate bucket.
   */

  for (hptr = table->buckets[index] ; hptr ; hptr = hptr->next_ptr) {
    for (p1 = key, p2 = hptr->key.string ; ; ++p1, ++p2) {
      if (*p1 != *p2) {
	break;
      }
      if (! *p1) {
	return hptr;
      }
    }
  }
  return NULL;
}

/*_____________________________________________________________________________
 *
 * string_create --
 *
 *	Given a hash table with string keys, and a string key, find
 *	the entry with a matching key.  If there is no matching entry,
 *	then create a new entry that does match.
 *
 * Results:
 *	The return value is a pointer to the matching entry.  If this
 *	is a newly-created entry, then *NEW will be set to a non-zero
 *	value;  otherwise *NEW will be set to 0.  If this is a new
 *	entry the value stored in the entry will initially be 0.
 *
 * Side effects:
 *	A new entry may be added to the hash table.
 *_____________________________________________________________________________
 */

static cp40_hash_entry_t *string_create(table, key, new)
     cp40_hash_table_t *table; /* Table in which to lookup entry. */
     const char *key;	       /* Key to use to find or create matching
				  entry. */
     int *new;	               /* Store info here telling whether a new
				  entry was created. */
{
  cp40_hash_entry_t *hptr;
  const char *name;
  size_t i, index;

  if (key == NULL) return NULL;
  index = hash_string((const unsigned char *)key) & table->mask;

  /* Search all of the entries in this bucket. */
  for (hptr = table->buckets[index] ; hptr ; hptr = hptr->next_ptr) {
    name = hptr->key.string;
    for (i = 0 ; key[i] == name[i] ; ++i) {
      if (! key[i]) {
	*new = 0;
	return hptr;
      }
    }
  }

  /* Entry not found.  Add a new one to the bucket. */
  *new = 1;
  hptr = (cp40_hash_entry_t *)malloc((unsigned)((sizeof(cp40_hash_entry_t)
						 - sizeof(hptr->key) + 1)
						+ strlen(key)));
  if (hptr == NULL) {
    cp40_error("not enough memory for new hash entry");
    return NULL;
  }
  hptr->table= table;
  hptr->bucket_ptr= &(table->buckets[index]);
  hptr->next_ptr= *hptr->bucket_ptr;
  hptr->client_data= 0;
  hptr->delete= NULL;
  strcpy(hptr->key.string, key);
  *hptr->bucket_ptr= hptr;
  table->num_entries++;

  /* If the table has exceeded a decent size, try to rebuild it with many more
     buckets (if rebuild_table fails it lefts the hash table unchanged). */
  if (table->num_entries >= table->rebuild_size) rebuild_table(table);
  return hptr;
}

/*_____________________________________________________________________________
 *
 * one_word_find --
 *
 *	Given a hash table with one-word keys, and a one-word key, find
 *	the entry with a matching key.
 *
 * Results:
 *	The return value is a token for the matching entry in the
 *	hash table, or NULL if there was no matching entry.
 *
 * Side effects:
 *	None.
 *_____________________________________________________________________________
 */

static cp40_hash_entry_t *one_word_find(table, key)
     cp40_hash_table_t *table;	        /* Table in which to lookup entry. */
     const char *key;	/* Key to use to find matching entry. */
{
  cp40_hash_entry_t *hptr;
  int index = RANDOM_INDEX(table, key);

  /* Search all of the entries in the appropriate bucket. */
  for (hptr=table->buckets[index] ; hptr != NULL ; hptr= hptr->next_ptr) {
    if (hptr->key.one_word_value == key) return hptr;
  }
  return NULL;
}

/*_____________________________________________________________________________
 *
 * one_word_create --
 *
 *	Given a hash table with one-word keys, and a one-word key, find
 *	the entry with a matching key.  If there is no matching entry,
 *	then create a new entry that does match.
 *
 * Results:
 *	The return value is a pointer to the matching entry.  If this
 *	is a newly-created entry, then *NEW will be set to a non-zero
 *	value;  otherwise *NEW will be set to 0.  If this is a new
 *	entry the value stored in the entry will initially be 0.
 *
 * Side effects:
 *	A new entry may be added to the hash table.
 *_____________________________________________________________________________
 */

static cp40_hash_entry_t *one_word_create(table, key, new)
     cp40_hash_table_t *table;        /* Table in which to lookup entry. */
     const char *key; /* Key to use to find or create matching
				  entry. */
     int *new;                 /* Store info here telling whether a new
				  entry was created. */
{
  cp40_hash_entry_t *hptr;
  int index= RANDOM_INDEX(table, key);

  /* Search all of the entries in this bucket. */
  for (hptr= table->buckets[index]; hptr != NULL; hptr= hptr->next_ptr) {
    if (hptr->key.one_word_value == key) {
      *new= 0;
      return hptr;
    }
  }

  /* Entry not found: add a new one to the bucket. */
  *new= 1;
  hptr= (cp40_hash_entry_t *)malloc(sizeof(cp40_hash_entry_t));
  if (hptr == NULL) {
    cp40_sys_error("cannot allocate memory for new hash entry");
    return NULL;
  }
  hptr->table= table;
  hptr->bucket_ptr= &(table->buckets[index]);
  hptr->next_ptr= *hptr->bucket_ptr;
  hptr->client_data= 0;
  hptr->delete= NULL;
  hptr->key.one_word_value= key;
  *hptr->bucket_ptr= hptr;
  table->num_entries++;

  /* If the table has exceeded a decent size, try to rebuild it with many more
     buckets (if rebuild_table fails it lefts the hash table unchanged). */
  if (table->num_entries >= table->rebuild_size) rebuild_table(table);
  return hptr;
}

/*_____________________________________________________________________________
 *
 * array_find --
 *
 *	Given a hash table with array-of-int keys, and a key, find
 *	the entry with a matching key.
 *
 * Results:
 *	The return value is a token for the matching entry in the
 *	hash table, or NULL if there was no matching entry.
 *
 * Side effects:
 *	None.
 *_____________________________________________________________________________
 */

static cp40_hash_entry_t *array_find(table, key)
    cp40_hash_table_t *table;	/* Table in which to lookup entry. */
    const char *key;		/* Key to use to find matching entry. */
{
    cp40_hash_entry_t *hptr;
    int *arrayPtr= (int *) key;
    int *iptr1, *iptr2;
    int index, count;

    for (index= 0, count= table->key_type, iptr1= arrayPtr;
	    count > 0; count--, iptr1++) {
	index += *iptr1;
    }
    index= RANDOM_INDEX(table, index);

    /*
     * Search all of the entries in the appropriate bucket.
     */

    for (hptr= table->buckets[index]; hptr != NULL;
	    hptr= hptr->next_ptr) {
	for (iptr1= arrayPtr, iptr2= hptr->key.words,
		count= table->key_type; ; count--, iptr1++, iptr2++) {
	    if (count == 0) {
		return hptr;
	    }
	    if (*iptr1 != *iptr2) {
		break;
	    }
	}
    }
    return NULL;
}

/*_____________________________________________________________________________
 *
 * array_create --
 *
 *	Given a hash table with one-word keys, and a one-word key, find
 *	the entry with a matching key.  If there is no matching entry,
 *	then create a new entry that does match.
 *
 * Results:
 *	The return value is a pointer to the matching entry.  If this
 *	is a newly-created entry, then *NEW will be set to a non-zero
 *	value;  otherwise *NEW will be set to 0.  If this is a new
 *	entry the value stored in the entry will initially be 0.
 *
 * Side effects:
 *	A new entry may be added to the hash table.
 *_____________________________________________________________________________
 */

static cp40_hash_entry_t *array_create(table, key, new)
     cp40_hash_table_t *table;	/* Table in which to lookup entry. */
     const char *key;	/* Key to use to find or create matching
				 * entry. */
     int *new;		/* Store info here telling whether a new
			 * entry was created. */
{
  cp40_hash_entry_t *hptr;
  int *arrayPtr= (int *) key;
  int *iptr1, *iptr2;
  int index, count;

  for (index= 0, count= table->key_type, iptr1= arrayPtr;
       count > 0; count--, iptr1++) {
    index += *iptr1;
  }
  index= RANDOM_INDEX(table, index);

  /*
   * Search all of the entries in the appropriate bucket.
   */

  for (hptr= table->buckets[index] ; hptr != NULL ; hptr= hptr->next_ptr) {
    for (iptr1= arrayPtr, iptr2= hptr->key.words,
	   count= table->key_type; ; count--, iptr1++, iptr2++) {
      if (count == 0) {
	*new= 0;
	return hptr;
      }
      if (*iptr1 != *iptr2) {
	break;
      }
    }
  }

  /*
   * Entry not found.  Add a new one to the bucket.
   */

  *new= 1;
  hptr= (cp40_hash_entry_t *)malloc((unsigned)(sizeof(cp40_hash_entry_t)
					 + (table->key_type*sizeof(int))
					 - 4));
  if (hptr == NULL) {
    cp40_error("not enough memory for new hash entry");
    return NULL;
  }
  hptr->table= table;
  hptr->bucket_ptr= &(table->buckets[index]);
  hptr->next_ptr= *hptr->bucket_ptr;
  hptr->client_data= 0;
  hptr->delete= NULL;
  for (iptr1=arrayPtr, iptr2=hptr->key.words, count=table->key_type;
       count > 0; count--, iptr1++, iptr2++) {
    *iptr2= *iptr1;
  }
  *hptr->bucket_ptr= hptr;
  table->num_entries++;

  /* If the table has exceeded a decent size, try to rebuild it with many more
     buckets (if rebuild_table fails it lefts the hash table unchanged). */
  if (table->num_entries >= table->rebuild_size) rebuild_table(table);
  return hptr;
}

/*_____________________________________________________________________________
 *
 * bogus_find --
 *
 *	This procedure is invoked when an cp40_hash_find_entry is called
 *	on a table that has been deleted.
 *
 * Results:
 *	If panic returns (which it shouldn't) this procedure returns
 *	NULL.
 *
 * Side effects:
 *	Generates a panic.
 *_____________________________________________________________________________
 */

/* ARGSUSED */
static cp40_hash_entry_t *bogus_find(table, key)
     cp40_hash_table_t *table;	/* Table in which to lookup entry. */
     const char *key;		/* Key to use to find matching entry. */
{
  cp40_fatal_error("called cp40_hash_find_entry on deleted table");
  return NULL;
}

/*_____________________________________________________________________________
 *
 * bogus_create --
 *
 *	This procedure is invoked when an cp40_hash_create_entry is called
 *	on a table that has been deleted.
 *
 * Results:
 *	If panic returns (which it shouldn't) this procedure returns
 *	NULL.
 *
 * Side effects:
 *	Generates a panic.
 *_____________________________________________________________________________
 */

/* ARGSUSED */
static cp40_hash_entry_t *bogus_create(table, key, new)
    cp40_hash_table_t *table;	/* Table in which to lookup entry. */
    const char *key;		/* Key to use to find or create matching
				 * entry. */
    int *new;		/* Store info here telling whether a new
			 * entry was created. */
{
  cp40_fatal_error("called cp40_hash_create_entry on deleted table");
  return NULL;
}

/*_____________________________________________________________________________
 *
 * rebuild_table --
 *
 *	This procedure is invoked when the ratio of entries to hash
 *	buckets becomes too large.  It creates a new table with a
 *	larger bucket array and moves all of the entries into the
 *	new table.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Memory gets reallocated and entries get re-hashed to new
 *	buckets.
 *
 * Bugs:
 *	Not robust against interruptions.
 *_____________________________________________________________________________
 */

static void rebuild_table(table)
     cp40_hash_table_t *table;	/* Table to enlarge. */
{
  int old_size, count, index;
  cp40_hash_entry_t **old_buckets;
  cp40_hash_entry_t **old_chain, **new_chain;
  cp40_hash_entry_t *hptr;

  old_size= table->num_buckets;
  old_buckets= table->buckets;

  /* Allocate and initialize the new bucket array, and set up hashing
     constants for new array size. */
  table->num_buckets *= 4;
  table->buckets= CP40_ALLOC_MANY(cp40_hash_entry_t *, table->num_buckets);
  if (table->buckets == NULL) {
    table->num_buckets= old_size;
    table->buckets= old_buckets;
    return;
  }

  for (count= table->num_buckets, new_chain= table->buckets;
       count > 0; count--, new_chain++) {
    *new_chain= NULL;
  }
  table->rebuild_size *= 4;
  table->down_shift -= 2;
  table->mask = (table->mask << 2) + 3;

  /* Rehash all of the existing entries into the new bucket array. */
  for (old_chain= old_buckets ; old_size > 0 ; old_size--, old_chain++) {
    for (hptr= *old_chain; hptr != NULL; hptr= *old_chain) {
      *old_chain= hptr->next_ptr;
      if (table->key_type == CP40_HASH_STRING_KEYS) {
	index= (hash_string((const unsigned char *)hptr->key.string)
		& table->mask);
      } else if (table->key_type == CP40_HASH_ONE_WORD_KEYS) {
	index= RANDOM_INDEX(table, hptr->key.one_word_value);
      } else {
	int *iptr= hptr->key.words;
	int n= table->key_type;
	for (index= 0 ; n > 0 ; n--, iptr++) {
	  index+= *iptr;
	}
	index= RANDOM_INDEX(table, index);
      }
      hptr->bucket_ptr= &(table->buckets[index]);
      hptr->next_ptr= *hptr->bucket_ptr;
      *hptr->bucket_ptr= hptr;
    }
  }

  /* Delete up the old bucket array, if it was dynamically allocated. */
  if (old_buckets != table->static_buckets) {
    free(old_buckets);
  }
}
