/*
 * cp40_low.c --
 *
 *      Implementation of most of the low-level interface to manage CP40 data
 *      attributes.
 *
 *-----------------------------------------------------------------------------
 *
 *	Copyright (C) 1995-2006 Eric Thibaut.
 *
 *	This file is part of CP40soft.
 *
 *	CP40soft is free software; you can redistribute it and/or modify it
 *	under the terms of the GNU General Public License as published by
 *	the Free Software Foundation; either version 2 of the License, or
 *	(at your option) any later version.
 *
 *	CP40soft is distributed in the hope that it will be useful, but
 *	WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *	General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with CP40soft; if not, write to the Free Software Foundation,
 *	Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
 *
 *-----------------------------------------------------------------------------
 *
 * History:
 *	$Id$
 *	$Log$
 */

#include <math.h>

extern int isnan(double);
extern int isinf(double);

#include "cp40.h"

/*---------------------------------------------------------------------------*/
/* PRIVATE MACROS */

#define DEG_2_RAD  0.01745329251994329576923690768488612713442871888541
#define ANGLE_IN_RADIANS(PTR) ((DEG_2_RAD)*((PTR)->angle))

/* Loop over each CP40 channels. */
#undef EACH_CHANNEL
#define EACH_CHANNEL(var) (var=0 ; var<CP40_NCHANNELS ; ++var)

/* Macros used to compute extreme coordinates after distortion correction. */
#if (CP40_NCOEFS != 10)
#error "only 3rd order polynomial correction supported"
#endif
#define XMIN (-0.5)
#define XMAX (CP40_CCD_WIDTH-0.5)
#define YMIN (-0.5)
#define YMAX (CP40_CCD_HEIGHT-0.5)

/* Macro to check channel number (causes routine to return with
   result CP40_ERROR if the channel number is invalid). */
#define CHECK_CHANNEL(c) if (cp40_check_channel(channel)!=CP40_OK) \
                           return CP40_ERROR

/*---------------------------------------------------------------------------*/

int cp40_check_channel(int channel)
{
  return ((unsigned int)channel < CP40_NCHANNELS) ? CP40_OK
    : cp40_error("bad channel number %d", channel);
}

/*---------------------------------------------------------------------------*/

/* FIXME: assume INSTANCE is valid (i.e. non-NULL). */

int cp40_get_nframes(cp40_instance_t *this)
{ return this ? this->nframes : -1; }

int cp40_get_nsaturations(cp40_instance_t *this)
{ return this ? this->nsaturations : -1; }

int cp40_get_nphotons(cp40_instance_t *this, int channel)
{
  if (! this || cp40_check_channel(channel) != CP40_OK) return -1;
  return this->nphotons[channel];
}
int cp40_get_frame_count(cp40_instance_t *this)
{ return this ? this->raw_frame->count : -1; }

int cp40_get_frame_date(cp40_instance_t *this)
{ return this ? this->raw_frame->date : -1; }

int cp40_get_exposure(cp40_instance_t *this)
{ return this ? (this->exposure ? this->exposure : 1) : -1; }

/*---------------------------------------------------------------------------*/

int cp40_set_scale(cp40_instance_t *this, double scale)
{
  if (this->task) {
    return cp40_error("cannot set \"scale\": scheduled tasks exist");
  }
  if (scale < 0.0 || isinf(scale) || isnan(scale)) {
    return cp40_error("bad value for \"scale\"");
  }
  if (this->scale != scale) {
    this->scale= scale;
    _cp40_invalidate_cache(this);
  }
  return CP40_OK;
}

double cp40_get_scale(cp40_instance_t *this)
{
  return this->scale;
}

/*---------------------------------------------------------------------------*/

int cp40_set_angle(cp40_instance_t *this, double angle)
{
  if (this->task) {
    return cp40_error("cannot set \"angle\": scheduled tasks exist");
  }
  if (isinf(angle) || isnan(angle)) {
    return cp40_error("bad value for \"angle\"");
  }
  angle= fmod(angle, 360.0);
  if (angle > 180.0) {
    angle -= 360.0;
  } else if (angle <= -180.0) {
    angle += 360.0;
  }
  if (this->angle != angle) {
    this->angle= angle;
    _cp40_invalidate_cache(this);
  }
  return CP40_OK;
}

double cp40_get_angle(cp40_instance_t *this)
{
  return this->angle;
}

/*---------------------------------------------------------------------------*/

extern int cp40_set_raw_coefs(cp40_instance_t *this, double *raw_coefs);
/* Query/set unscaled polynomial coefficients for distortion correction. */

int cp40_set_raw_coefs(cp40_instance_t *this, double *raw_coefs)
{
  int i;

  if (this->task) {
    return cp40_error("cannot set \"raw_coefs\": scheduled tasks exist");
  }

  for (i=0 ; i<CP40_COEF_COUNT ; ++i) {
    if (isnan(raw_coefs[i]) || isinf(raw_coefs[i])) {
      return cp40_error("bad value for \"raw_coefs\"");
    }
  }
  for (i=0 ; i<CP40_COEF_COUNT ; ++i) {
    this->raw_coefs[i]= raw_coefs[i];
  }
  _cp40_invalidate_cache(this);
  return CP40_OK;
}

void cp40_get_raw_coefs(cp40_instance_t *this, double *raw_coefs)
{
  int i;
  for (i=0 ; i<CP40_COEF_COUNT ; ++i) raw_coefs[i]= this->raw_coefs[i];
}

void cp40_get_coefs(cp40_instance_t *this, double *coefs)
{
  int i;
  CP40_UPDATE_CACHE_MAYBE(this);
  for (i=0 ; i<CP40_COEF_COUNT ; ++i) coefs[i]= this->coefs[i];
}

/*---------------------------------------------------------------------------*/

int cp40_set_pixel(cp40_instance_t *this, cp40_pixel_t value)
{
  if (this->task) {
    return cp40_error("cannot set \"pixel\": scheduled tasks exist");
  }
  switch (value) {
  case CP40_PIXEL_LOGICAL:
  case CP40_PIXEL_HALF_PHYSICAL:
  case CP40_PIXEL_PHYSICAL:
    if (this->pixel != value) {
      this->pixel= value;
      _cp40_invalidate_cache(this);
    }
    return CP40_OK;
  }
  return cp40_error("bad value for \"pixel\"");
}

cp40_pixel_t cp40_get_pixel(cp40_instance_t *this)
{
  return this->pixel;
}

/*---------------------------------------------------------------------------*/

int cp40_set_parity(cp40_instance_t *this, int parity)
{
  switch (parity) {
  case CP40_PARITY_EVEN:
  case CP40_PARITY_ODD:
  case CP40_PARITY_ANY:
    this->parity = parity;
    return CP40_OK;
  }
  return cp40_error("bad value for \"parity\"");
}

int cp40_get_parity(cp40_instance_t *this)
{
  return this->parity;
}

/*---------------------------------------------------------------------------*/

static int range_copy(cp40_range_t *dst, cp40_range_t *src, int check);

#undef GET_FUNCTION
#define GET_FUNCTION(MEMBER)						      \
int CP40_JOIN(cp40_get_,MEMBER)(cp40_instance_t *this, int channel,	      \
			     cp40_range_t *range_ptr)			      \
{									      \
  CHECK_CHANNEL(channel);						      \
  return range_copy(range_ptr, &this->MEMBER[channel], 0);		      \
}

#undef SET_FUNCTION
#define SET_FUNCTION(MEMBER)						      \
int CP40_JOIN(cp40_set_,MEMBER)(cp40_instance_t *this, int channel,	      \
			     cp40_range_t *range_ptr)			      \
{									      \
  CHECK_CHANNEL(channel);						      \
  return range_copy(&this->MEMBER[channel], range_ptr, 1);		      \
}

GET_FUNCTION(flux)
SET_FUNCTION(flux)
GET_FUNCTION(spread)
SET_FUNCTION(spread)
GET_FUNCTION(distance)
SET_FUNCTION(distance)
GET_FUNCTION(ratio)
SET_FUNCTION(ratio)

static int range_copy(cp40_range_t *dst, cp40_range_t *src, int check)
{
  if (check) {
    if (src->flags & ~(CP40_MIN_SET | CP40_MAX_SET)) {
      return cp40_error("bad range flags");
    }
    if ((src->flags & CP40_MIN_SET) && src->min < 0.0) {
      return cp40_error("bad range (min < 0)");
    }
    if ((src->flags & CP40_MAX_SET) && src->max < 0.0) {
      return cp40_error("bad range (max < 0)");
    }
    if ((src->flags & CP40_MIN_SET) && (src->flags & CP40_MAX_SET)
	&& src->min > src->max) {
      return cp40_error("bad range (min > max)");
    }
  }
  if (dst != NULL && dst != src) {
    dst->flags= src->flags;
    dst->min= src->min;
    dst->max= src->max;
  }
  return CP40_OK;
}

/*---------------------------------------------------------------------------*/

#undef GET_FUNCTION
#define GET_FUNCTION(MEMBER)						      \
int CP40_JOIN(cp40_get_,MEMBER)(cp40_instance_t *this, int channel,	      \
			       cp40_window_t *win)			      \
{									      \
  CHECK_CHANNEL(channel);						      \
  CP40_UPDATE_CACHE_MAYBE(this);					      \
  win->xmin = this->MEMBER[channel].xmin;				      \
  win->xmax = this->MEMBER[channel].xmax;				      \
  win->ymin = this->MEMBER[channel].ymin;				      \
  win->ymax = this->MEMBER[channel].ymax;				      \
  return CP40_OK;							      \
}

int cp40_set_window(cp40_instance_t *this, int channel, cp40_window_t *win)
{
  cp40_window_t *dst;
  CHECK_CHANNEL(channel);
  if (this->task) {
    return cp40_error("cannot set \"window\": scheduled tasks exist");
  }

  /* only copy window if it is different from old one */
  dst = &this->window[channel];
  if (dst != win && (dst->xmin != win->xmin || dst->xmax != win->xmax ||
		     dst->ymin != win->ymin || dst->ymax != win->ymax)) {
    /* check input window boundaries */
    if (win->xmin != CP40_WINDOW_WILD &&
	win->xmin < 0) return cp40_error("bad window (xmin < 0)");
    if (win->xmax != CP40_WINDOW_WILD) {
      if (win->xmax < 0) return cp40_error("bad window (xmax < 0)");
      if (win->xmin != CP40_WINDOW_WILD &&
	  win->xmin > win->xmax) return cp40_error("bad window (xmin > xmax)");
    }
    if (win->ymin != CP40_WINDOW_WILD &&
	win->ymin < 0) return cp40_error("bad window (ymin < 0)");
    if (win->ymax != CP40_WINDOW_WILD) {
      if (win->ymax < 0) return cp40_error("bad window (ymax < 0)");
      if (win->ymin != CP40_WINDOW_WILD &&
	  win->ymin > win->ymax) return cp40_error("bad window (ymin > ymax)");
    }

    /* update window settings and invalidate cache */
    dst->xmin= win->xmin;
    dst->xmax= win->xmax;
    dst->ymin= win->ymin;
    dst->ymax= win->ymax;
    _cp40_invalidate_cache(this);
  }
  return CP40_OK;
}
GET_FUNCTION(extreme)
GET_FUNCTION(window)

/*---------------------------------------------------------------------------*/

int cp40_get_limits(cp40_instance_t *this, int channel,
		    int *xmin, int *xmax, int *ymin, int *ymax)
{
  cp40_frame_t *frame;

  CHECK_CHANNEL(channel);
  CP40_UPDATE_CACHE_MAYBE(this);
  frame= &this->frame[channel];
  *xmin= frame->xmin;
  *xmax= frame->xmax;
  *ymin= frame->ymin;
  *ymax= frame->ymax;
  return CP40_OK;
}

int cp40_get_limits2(cp40_instance_t *this, int channel1, int channel2,
		     int *xmin_ptr, int *xmax_ptr,
		     int *ymin_ptr, int *ymax_ptr)
{
  int xmin1, xmax1, ymin1, ymax1;
  int xmin2, xmax2, ymin2, ymax2;

  if (cp40_get_limits(this, channel1,
		      &xmin1, &xmax1, &ymin1, &ymax1) != CP40_OK) {
    return CP40_ERROR;
  }
  if (channel1 == channel2) {
    *xmin_ptr= xmin1 - xmax1;
    *xmax_ptr= xmax1 - xmin1;
    *ymin_ptr= ymin1 - ymax1;
    *ymax_ptr= ymax1 - ymin1;
  } else {
    if (cp40_get_limits(this, channel2,
			&xmin2, &xmax2, &ymin2, &ymax2) != CP40_OK) {
      return CP40_ERROR;
    }
    *xmin_ptr= xmin2 - xmax1;
    *xmax_ptr= xmax2 - xmin1;
    *ymin_ptr= ymin2 - ymax1;
    *ymax_ptr= ymax2 - ymin1;
  }
  return CP40_OK;
}

/*---------------------------------------------------------------------------*/

#if 0
static void set_default_correction(cp40_instance_t *this)
{
  static double raw_coefs[]= {
    2.2429018895094273, 0.022351366698291554,
    -0.0028881561337462682, 4.4258639834761686e-05,
    2.3729185234885372e-05, 1.8674411322662041e-05,
    -3.9226788987704291e-08, 5.8756878741836713e-09,
    -3.9825549564531035e-08, -2.8159900257027333e-09,
    4.5564864158259546, -0.012565638643869548,
    0.028217046485032132, -1.6691140625571879e-07,
    2.2443944809214089e-05, 3.3736508460210937e-05,
    1.5935015678147993e-08, -4.9194493672466417e-08,
    2.4920095646503966e-08, -4.318678985377763e-08,
    7.5368875448424895, 0.038425434416869901,
    0.0060963331423668039, -2.6903728765339331e-06,
    3.1243764326070336e-05, 1.1556173258962149e-05,
    -3.2568836700622905e-08, -2.7140701079990379e-08,
    -4.489559716490471e-08, -1.5860460326349369e-08,
    0.88732931342433041, -0.0084599377132116155,
    0.029298002406205958, 1.1151469977053874e-05,
    -3.1304331517921581e-06, 4.1546606895317568e-05,
    1.3703578447542688e-08, -3.9191942070376493e-08,
    -8.0203994879539308e-09, -3.9630495355515094e-08,
    5.3442574889629144, 0.037907183811540561,
    0.0046726801173435726, 5.2706304983631377e-06,
    3.5221941633501079e-05, 1.2773746946483091e-05,
    -4.3189285239403624e-08, -2.5267416265154023e-08,
    -5.141288403943272e-08, -1.2540514037695845e-08,
    3.9584265117677604, -0.010244613657504363,
    0.030208911468528596, 1.6654293814152014e-05,
    -2.9917509072072171e-06, 4.030415030038905e-05,
    2.9338123825014282e-09, -3.6046579580381278e-08,
    -4.2005130647209745e-09, -3.7360440477251416e-08,
    6.526683656862085, 0.019379115399177547,
    -0.0041467229292821159, 4.6867622530945166e-05,
    2.4337192886464773e-05, 2.3723712313582782e-05,
    -3.8323625457377353e-08, 1.0175097258487054e-08,
    -3.9964094965111904e-08, -5.3356128511457242e-09,
    2.5892577491433308, -0.013447959988497047,
    0.027689096380291371, -3.3450240783578899e-07,
    2.4198777435786562e-05, 3.3223297823376655e-05,
    1.4196135956156128e-08, -5.4700034168284679e-08,
    3.6382906516489581e-08, -4.611782006595048e-08,
  };
  if (this->task) {
    return cp40_error("cannot set \"raw_coefs\": scheduled tasks exist");
  }
  memcpy(this->raw_coefs, raw_coefs, sizeof(raw_coefs));
  this->notify= 1;
}
#endif

/*---------------------------------------------------------------------------*/

int cp40_shift(cp40_instance_t *this, int channel,
	       double dx, double dy)
{
  double a, b, *raw_coefs;

  if (this->task) {
    return cp40_error("cannot fix offsets: scheduled tasks exist");
  }
  if (cp40_check_channel(channel) != CP40_OK) return CP40_ERROR;
  if (this->scale <= 0.0 ||
      cp40_get_correction(this) == CP40_CORRECTION_NONE) {
    return cp40_error("cannot fix offsets: no distortion correction");
  }

  /* De-magnify and de-rotate offsets before shifting
     degree zero raw coefficients. */
  a = -ANGLE_IN_RADIANS(this);
  b = sin(a)/this->scale;
  a = cos(a)/this->scale;
  raw_coefs = &this->raw_coefs[channel*(2*CP40_NCOEFS)];
  raw_coefs[0]           += a * dx - b * dy;
  raw_coefs[CP40_NCOEFS] += b * dx + a * dy;

  _cp40_invalidate_cache(this);  /* distortion correction coefficients
				    must be re-computed */
  return CP40_OK;
}

/*---------------------------------------------------------------------------*/

/* Compute extreme positions after distortion correction. */
static void get_extreme(double *coefs, double x, double y,
			double *xmin, double *xmax,
			double *ymin, double *ymax);


void cp40_update_cache(cp40_instance_t *this)
{
  int i, ix, iy, channel;
  double xmin[4], xmax[4], ymin[4], ymax[4];
  double xoffset, yoffset;
  double a, b;
  double *coefs, *raw_coefs;
  cp40_correction_t correction;

  /* Basic checking. */
  if (this->notify==0) return; /* nothing to do */
  if (this->task) {
    cp40_warning("cannot update settings: scheduled tasks exist");
    return;
  }

  /* Fix distortion correction method. */
  correction= cp40_get_correction(this);
  if (this->scale<=0.0) {
    this->scale= 0.0;
    if (correction != CP40_CORRECTION_NONE) {
      cp40_warning("no distortion correction with SCALE<=0");
      correction= CP40_CORRECTION_NONE;
      cp40_set_correction(this, correction);
    }
  }

  if (correction==CP40_CORRECTION_NONE) {

    /* Set settings of distortion correction so that distortion correction
       is equivalent to identity.   Just in case someone query these
       coefficients... */

    int raw_xmax= CP40_WIDTH(this->pixel) - 1;
    int raw_ymax= CP40_HEIGHT(this->pixel) - 1;
    coefs= this->coefs;
    for EACH_CHANNEL(channel) {
      for (i=0 ; i<CP40_NCOEFS ; ++i) *coefs++ = i==1 ? 1.0 : 0.0;
      for (i=0 ; i<CP40_NCOEFS ; ++i) *coefs++ = i==2 ? 1.0 : 0.0;
      this->extreme[channel].xmin= 0;
      this->extreme[channel].xmax= raw_xmax;
      this->extreme[channel].ymin= 0;
      this->extreme[channel].ymax= raw_ymax;
    }

  } else {

    /* Distortion correction settings must be recomputed.  For each channel,
       compute effective coefficients of distortion correction and extreme
       positions. */

    for EACH_CHANNEL(channel) {

      /* Coefficients for current channel. */
      raw_coefs= &this->raw_coefs[channel * (2 * CP40_NCOEFS)];
      coefs= &this->coefs[channel * (2 * CP40_NCOEFS)];

      /* Apply rotation and magnification: magnification=THIS->scale
	 and rotation=THIS->angle (or THIS->angle + 180 deg. for
	 channels 2 and 3). */
      a= ANGLE_IN_RADIANS(this);
      b= this->scale * sin(a);
      a= this->scale * cos(a);
      for (ix=0, iy=CP40_NCOEFS ; ix<CP40_NCOEFS ; ++ix, ++iy) {
	coefs[ix]= a * raw_coefs[ix] - b * raw_coefs[iy];
	coefs[iy]= b * raw_coefs[ix] + a * raw_coefs[iy];
      }

      /* Seek for the extrema of coordinates after correction (at the
	 corners, and then along the edges) and add a small margin to these
	 extrema to be sure that rounding errors will not give positions
	 outside these boundaries. */
      xmin[channel]= xmax[channel]= coefs[0];
      ymin[channel]= ymax[channel]= coefs[CP40_NCOEFS];
#define GET_MIN_MAX(x, y) get_extreme(coefs, x, y, \
	&xmin[channel], &xmax[channel], &ymin[channel], &ymax[channel])
      GET_MIN_MAX(XMIN, YMIN);
      GET_MIN_MAX(XMIN, YMAX);
      GET_MIN_MAX(XMAX, YMIN);
      GET_MIN_MAX(XMAX, YMAX);
      GET_MIN_MAX(XMIN, -1);
      GET_MIN_MAX(XMAX, -1);
      GET_MIN_MAX(-1, YMIN);
      GET_MIN_MAX(-1, YMAX);
    }

#if 0 /* NO NEED FOR THAT EXTRA BOUND BECAUSE ALL PHOTON POSITIONS
	 ARE CHECKED AGAINST WINDOW SETTINGS. */

    /* Add a small safety boundary around the limits found. */
    a= this->scale / 1e6;
    for EACH_CHANNEL(channel) {
      xmin[channel]-= a;
      xmax[channel]+= a;
      ymin[channel]-= a;
      ymax[channel]+= a;
    }
#endif

    /* Compute and apply global offsets to ensure that coordinates are
       always positive. */
    xoffset= xmin[0];
    yoffset= ymin[0];
    for (channel=1 ; channel<CP40_NCHANNELS ; ++channel) {
      if (xmin[channel] < xoffset) xoffset= xmin[channel];
      if (ymin[channel] < yoffset) yoffset= ymin[channel];
    }
    for (i=-CP40_NCOEFS, channel=0 ; channel<CP40_NCHANNELS ; ++channel) {
      this->coefs[i+= CP40_NCOEFS]-= xoffset;
      this->coefs[i+= CP40_NCOEFS]-= yoffset;
      this->extreme[channel].xmin= floor(xmin[channel] - xoffset);
      this->extreme[channel].xmax=  ceil(xmax[channel] - xoffset);
      this->extreme[channel].ymin= floor(ymin[channel] - yoffset);
      this->extreme[channel].ymax=  ceil(ymax[channel] - yoffset);
    }

    /* Ultimate rescaling of coefficients to account for that distorted
       positions may not be in physical CCD pixels (as assumed by raw
       coefficients). */
#if CP40_NCOEFS != 10
# error "following C-code assumes CP40_NCOEFS=10"
#endif
    if (this->pixel != CP40_PIXEL_PHYSICAL) {
      a= (this->pixel == CP40_PIXEL_LOGICAL) ? 4.0 : 2.0;
      for EACH_CHANNEL(channel) {
	coefs= &this->coefs[channel * (2 * CP40_NCOEFS)];
	b= a;
	coefs[1]/= b; coefs[11]/= b;
	coefs[2]/= b; coefs[12]/= b;
	b*= a;
	coefs[3]/= b; coefs[13]/= b;
	coefs[4]/= b; coefs[14]/= b;
	coefs[5]/= b; coefs[15]/= b;
	b*= a;
	coefs[6]/= b; coefs[16]/= b;
	coefs[7]/= b; coefs[17]/= b;
	coefs[8]/= b; coefs[18]/= b;
	coefs[9]/= b; coefs[19]/= b;
      }
    }
  }

  /* Update frame boundaries. */
  for EACH_CHANNEL(channel) {
    cp40_window_t *wp= &this->window[channel];
    cp40_window_t *ep= &this->extreme[channel];
    cp40_frame_t  *fp= &this->frame[channel];
    fp->xmin= wp->xmin!=CP40_WINDOW_WILD ? wp->xmin : ep->xmin;
    fp->xmax= wp->xmax!=CP40_WINDOW_WILD ? wp->xmax : ep->xmax;
    fp->ymin= wp->ymin!=CP40_WINDOW_WILD ? wp->ymin : ep->ymin;
    fp->ymax= wp->ymax!=CP40_WINDOW_WILD ? wp->ymax : ep->ymax;
  }

  /* All settings are now up to date. */
  this->notify= 0;
}

/*---------------------------------------------------------------------------*/

static int poly_root(int deg, const double a[], double x[]);


#undef SET_MIN_MAX
#define SET_MIN_MAX(min,max,x) if (x<min) min=x; else if (x>max) max=x

static void get_extreme(double *coefs,/* Distortion correction coefficients. */
			double x,     /* Abscissa to look at, -1 means any. */
			double y,     /* Ordinate to look at, -1 means any. */
			double *xmin, /* Address of minimum abscissa. */
			double *xmax, /* Address of maximum abscissa. */
			double *ymin, /* Address of minimum ordinate. */
			double *ymax) /* Address of maximum ordinate. */
{
  double c[3], root[2];
  double tmp;
  int n, nrac;

  if (x == -1) {

    /* Seek for extrema, Y being constant. */

    c[0]= coefs[1] + y * (coefs[4] + y * coefs[8]);
    c[1]= 2.0 * (coefs[3] + y * coefs[7]);
    c[2]= 3.0 * coefs[6];
    nrac= poly_root(2, c, root);
    for (n= 0; n < nrac; n++) {
      if ((tmp= root[n]) >= XMIN && tmp <= XMAX) {
	tmp= CP40_XCOR(coefs, tmp, y);
	SET_MIN_MAX(*xmin, *xmax, tmp);
      }
    }
    c[0]= coefs[10 + 1] + y * (coefs[10 + 4] + y * coefs[10 + 8]);
    c[1]= 2.0 * (coefs[10 + 3] + y * coefs[10 + 7]);
    c[2]= 3.0 * coefs[10 + 6];
    nrac= poly_root(2, c, root);
    for (n= 0; n < nrac; n++) {
      if ((tmp= root[n]) >= XMIN && tmp <= XMAX) {
	tmp= CP40_YCOR(coefs, tmp, y);
	SET_MIN_MAX(*ymin, *ymax, tmp);
      }
    }

  } else if (y == -1) {

    /* Seek for extrema, X being constant. */

    c[0]= coefs[2] + x * (coefs[4] + x * coefs[7]);
    c[1]= 2.0 * (coefs[5] + x * coefs[8]);
    c[2]= 3.0 * coefs[9];
    nrac= poly_root(2, c, root);
    for (n= 0; n < nrac; n++) {
      if ((tmp= root[n]) >= YMIN && tmp <= YMAX) {
	tmp= CP40_XCOR(coefs, x, tmp);
	SET_MIN_MAX(*xmin, *xmax, tmp);
      }
    }
    c[0]= coefs[10 + 2] + x * (coefs[10 + 4] + x * coefs[10 + 7]);
    c[1]= 2.0 * (coefs[10 + 5] + x * coefs[10 + 8]);
    c[2]= 3.0 * coefs[10 + 9];
    nrac= poly_root(2, c, root);
    for (n= 0; n < nrac; n++) {
      if ((tmp= root[n]) >= YMIN && tmp <= YMAX) {
	tmp= CP40_YCOR(coefs, x, tmp);
	SET_MIN_MAX(*ymin, *ymax, tmp);
      }
    }

  } else {

    /* Seek for extrema, X and Y being constant. */
    tmp= CP40_XCOR(coefs, x, y);
    SET_MIN_MAX(*xmin, *xmax, tmp);
    tmp= CP40_YCOR(coefs, x, y);
    SET_MIN_MAX(*ymin, *ymax, tmp);
  }
}

/*---------------------------------------------------------------------------*/
/*_____________________________________________________________________________
 *
 * poly_root --
 *
 *	Get real roots of a polynomial (of maximum degree 3), i.e. compute
 *	the real solution of:
 *	    A[0] + A[1] * X + A[2] * X^2 + ... + A[N] * X^N = 0
 *
 * Returns:
 *	Number M of real roots; -1 if there are an infinity of roots; -2
 *	in case of error.
 *
 * Side effects:
 *	Roots are stored into X[0], ..., X[M-1] where M is the number of
 *	roots.
 *_____________________________________________________________________________
 */

static int poly_root(int deg, const double a[], double x[])
     /* DEG= degree of polynomial
	A= array of DEG+1 polynomial coefficients
	X= array of (at least) DEG output roots. */
{
  /* Check arguments */
  if (deg < 0) {
    (void)cp40_error("invalid polynomial degree: %d", deg);
    return -2;
  }
  /* Reduction of polynomial order */
  while (deg >= 0 && a[deg] == 0.0) {
    deg--;
  }
  if (deg == -1) {
    return -1;		/* every x is solution (all the coefs. are zero) */
  }
  if (a[0] == 0.0) {
    x[0] = 0.0;
    return 1 + poly_root(deg - 1, a + 1, x + 1);
  }
  switch (deg) {
  case 0:
    return 0;			/* no solution */
  case 1:
    x[0] = -a[0] / a[1];
    return 1;			/* 1 solution */
  case 2:{
      double delta = a[1] * a[1] - 4.0 * a[0] * a[2];
      double q;

      if (delta > 0.0) {
	q = -0.5 * (a[1] >= 0 ? a[1] + sqrt(delta) : a[1] - sqrt(delta));
	x[0] = q / a[2];
	x[1] = a[0] / q;
	return 2;
      } else if (delta == 0.0) {
	x[0] = -0.5 * a[1] / a[2];
	return 1;
      } else {			/* delta < 0.0 */
	return 0;
      }
    }
  case 3:{
      double a0 = a[0] / a[3];
      double a1 = a[1] / a[3];
      double a2 = a[2] / (3.0 * a[3]);
      double q = a2 * a2 - a1 / 3.0;
      double r = a2 * (a2 * a2 - 0.5 * a1) + 0.5 * a0;
      double p = q * q * q - r * r;
      if (p > 0.0) {
	q = sqrt(q);
	p = acos(r / q / q / q) / 3.0;
	x[0] = -2.0 * q * cos(p) - a2;
	x[1] = -2.0 * q * cos(p + 2.09439510239319549230) - a2;
	x[2] = -2.0 * q * cos(p + 4.18879020478639098461) - a2;
	return 3;
      } else if (p == 0.0) {
	q = sqrt(q);
	x[0] = -2.0 * q - a2;
	x[2] = x[1] = q - a2;
	return 3;
      } else if (r > 0.0) {
	p = pow(sqrt(-p) + r, 1.0 / 3.0);
	x[0] = -p - q / p - a2;
      } else if (r < 0.0) {
	p = pow(sqrt(-p) - r, 1.0 / 3.0);
	x[0] = p + q / p - a2;
      } else {
	x[0] = -a2;
      }
      return 1;
    }
  default:
    (void)cp40_error("unsupported polynomial degree: %d", deg);
    return -2;
  }
}

/*---------------------------------------------------------------------------*/
