/*
 * cp40_xdr.c --
 *
 *	XDR (eXternal Data Representation) input/output routines for CP40.
 *	This package can be compiled alone (i.e. out of CP40).
 *
 * Warning:
 *	This package may not be portable since it assumes that at most byte
 *	swapping is needed for your machine data layout to match XDR (works
 *	on most IEEE machines: PC's, Sun, IBM or HP workstations, ...).
 *	You may also have to modify some typedefs...
 *
 *-----------------------------------------------------------------------------
 *
 *	Copyright (C) 1995-2006 Eric Thibaut.
 *
 *	This file is part of CP40soft.
 *
 *	CP40soft is free software; you can redistribute it and/or modify it
 *	under the terms of the GNU General Public License as published by
 *	the Free Software Foundation; either version 2 of the License, or
 *	(at your option) any later version.
 *
 *	CP40soft is distributed in the hope that it will be useful, but
 *	WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *	General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with CP40soft; if not, write to the Free Software Foundation,
 *	Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
 *
 *-----------------------------------------------------------------------------
 *
 * History:
 *	$Id$
 *	$Log$
 */

#ifndef _CP40_XDR_C
#define _CP40_XDR_C 1

#include <string.h>

#include "cp40.h"

/*---------------------------------------------------------------------------*/

#undef cp40_xdr_read_byte
size_t cp40_xdr_read_byte(void *ptr, size_t number, FILE *stream)
{
  return fread(ptr, 1, number, stream);
}

#undef cp40_xdr_write_byte
size_t cp40_xdr_write_byte(const void *ptr, size_t number, FILE *stream)
{
  return fwrite(ptr, 1, number, stream);
}

#define SWAP_BUF_SIZE 1024
static unsigned char swap_buf[SWAP_BUF_SIZE];
static int byte_order = 0;


int cp40_get_byte_order(void)
{
  if (! byte_order) {
    size_t i;
    cp40_uint32_t value = 0UL;
    char *ptr = (char *)&value;
    if (sizeof(cp40_uint16_t) != 2) {
      cp40_fatal_error("*** FATAL ERROR: sizeof(cp40_uint16_t) != 2");
    }
    if (sizeof(cp40_uint32_t) != 4) {
      cp40_fatal_error("*** FATAL ERROR: sizeof(cp40_uint32_t) != 4");
    }
    for (i=0 ; i<sizeof(value) ; ++i) {
      ptr[i] = 4 - i%4;
    }
    value &= 0xffffffff;
    if (value == 0x01020304) {
      byte_order = CP40_LITTLE_ENDIAN;
    } else if (value == 0x04030201) {
      byte_order = CP40_BIG_ENDIAN;
#if 0
    } else if (value == 0x03040102) {
      byte_order = CP40_PDP_ENDIAN;
#endif
    } else {
      cp40_fatal_error("*** FATAL ERROR: unsupported byte order");
    }
  }
  return byte_order;
}

/*---------------------------------------------------------------------------*/

int cp40_swap_bytes(void *dst, const void *src,
		    size_t elsize, size_t number)
{
  size_t i;

#undef SWAP
#define SWAP(i,j) c=p[i];p[i]=p[j];p[j]=c

  if (dst != src) memcpy(dst, src, elsize*number);
  switch (elsize) {
  case 1:
    break;
  case 2:
    {
      const cp40_uint16_t msk16_1 = 0x00ffU;
      const cp40_uint16_t msk16_2 = 0xff00U;
      cp40_uint16_t val, *p = dst;
      for (i=0 ; i<number ; ++i) {
	val = p[i];
	p[i] = (((val>>8) & msk16_1) | ((val<<8) & msk16_2));
      }
    }
    break;
  case 4:
    {
      const cp40_uint32_t msk32_1 = 0x000000ffU;
      const cp40_uint32_t msk32_2 = 0x0000ff00U;
      const cp40_uint32_t msk32_3 = 0x00ff0000U;
      const cp40_uint32_t msk32_4 = 0xff000000U;
      cp40_uint32_t val, *p = dst;
      for (i=0 ; i<number ; ++i) {
	val = p[i];
	p[i]= (((val>>24) & msk32_1) | ((val>> 8) & msk32_2) |
	       ((val<< 8) & msk32_3) | ((val<<24) & msk32_4));
      }
    }
    break;
  case 8:
    {
      unsigned char c, *p = (unsigned char *)dst;
      for (i=0 ; i<number ; ++i, p+=8) {
	SWAP(0,7);
	SWAP(1,6);
	SWAP(2,5);
	SWAP(3,4);
      }
    }
    break;
  case 16:
    {
      unsigned char c, *p = (unsigned char *)dst;
      for (i=0 ; i<number ; ++i, p+=8) {
	SWAP(0,15);
	SWAP(1,14);
	SWAP(2,13);
	SWAP(3,12);
	SWAP(4,11);
	SWAP(5,10);
	SWAP(6, 9);
	SWAP(7, 8);
      }
    }
    break;
  default:
    return cp40_error("cannot swap %lu byte data", (unsigned long)elsize);
  }
  return CP40_OK;
}

/*---------------------------------------------------------------------------*/

#undef XDR_SIZE
#undef XDR_READ
#undef XDR_WRITE

#define XDR_SIZE  2
#define XDR_READ  cp40_xdr_read_short
#define XDR_WRITE cp40_xdr_write_short
#include __FILE__

#define XDR_SIZE  4
#define XDR_READ  cp40_xdr_read_long
#define XDR_WRITE cp40_xdr_write_long
#include __FILE__

#define XDR_SIZE  4
#define XDR_READ  cp40_xdr_read_float
#define XDR_WRITE cp40_xdr_write_float
#include __FILE__

#define XDR_SIZE  8
#define XDR_READ  cp40_xdr_read_double
#define XDR_WRITE cp40_xdr_write_double
#include __FILE__

/*---------------------------------------------------------------------------*/
#else /* _CP40_XDR_C defined */

#undef SWAP
#define SWAP(i,j) c=data[i];data[i]=data[j];data[j]=c

size_t XDR_READ(void *ptr, size_t number, FILE *stream)
{
  if (! byte_order) cp40_get_byte_order();
  if ((number = fread(ptr, XDR_SIZE, number, stream)) > 0 &&
      byte_order == CP40_LITTLE_ENDIAN) {
    register size_t i, n = number*XDR_SIZE;
    register unsigned char c, *data = (unsigned char *)ptr;
    for (i=0 ; i<n ; i+=XDR_SIZE) {
#if XDR_SIZE==2
      SWAP(i, i+1);
#endif
#if XDR_SIZE==4
      SWAP(i,   i+3);
      SWAP(i+1, i+2);
#endif
#if XDR_SIZE==8
      SWAP(i,   i+7);
      SWAP(i+1, i+6);
      SWAP(i+2, i+5);
      SWAP(i+3, i+4);
#endif
    }
  }
  return number;
}

size_t XDR_WRITE(const void *ptr, size_t  number, FILE *stream)
{
  size_t n, nwrite, nbuf, nleft;
  unsigned char *dst;
  const unsigned char *src;

  /* May be no swap needed. */
  if (number == 0) return 0;

  /* Swap bytes into static buffer, then write bytes. */
  nbuf = (SWAP_BUF_SIZE / XDR_SIZE);
  src = ptr;
  for (nleft = number; nleft > 0; nleft -= nwrite) {
    dst = swap_buf;
    nwrite = (nleft < nbuf ? nleft : nbuf);
    for (n=0; n<nwrite; n++, src+=XDR_SIZE, dst+=XDR_SIZE) {
#if XDR_SIZE == 2
      dst[0] = src[1];
      dst[1] = src[0];
#elif XDR_SIZE == 4
      dst[0] = src[3];
      dst[1] = src[2];
      dst[2] = src[1];
      dst[3] = src[0];
#elif XDR_SIZE == 8
      dst[0] = src[7];
      dst[1] = src[6];
      dst[2] = src[5];
      dst[3] = src[4];
      dst[4] = src[3];
      dst[5] = src[2];
      dst[6] = src[1];
      dst[7] = src[0];
#endif
    }
    if ((n = fwrite(swap_buf, XDR_SIZE, nwrite, stream)) != nwrite) {
      if (n > 0) nleft -= n;
      break;
    }
  }
  return number - nleft;
}

#undef XDR_WRITE
#undef XDR_READ
#undef XDR_SIZE
/*---------------------------------------------------------------------------*/
#endif /* _CP40_XDR_C */
